#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.

from typing import Any, Callable, Dict, List, OrderedDict
from weakref import ReferenceType

from cosapp.core.module import Module
from cosapp.ports.port import Port
from cosapp.systems import System
from cosapp.tools.help import DocDisplay
from cosapp_lab.widgets.base.base_component import BaseComponent
from cosapp_lab.widgets.utils import CosappObjectParser


class SystemInfoComponent(BaseComponent):
    name = "SystemInfoComponent"

    def __init__(
        self,
        data: "ReferenceType[System]",
        sys_data: CosappObjectParser,
        send_func: Callable,
        **kwargs,
    ):
        super().__init__(data=data, sys_data=sys_data, send_func=send_func, **kwargs)


    @staticmethod
    def create_html(sys_obj : System, context_name: str) -> str:
        """Returns the markdown-formatted documentation of the input system.

        Parameters
        ----------
        sys_obj: System
            System object to generate documentation

        context_name: str
            Current contextual name of input object.
            
        Returns
        -------
        str
            String in Markdown format.
        """

        raw_string = DocDisplay(sys_obj)._repr_markdown_().strip()
        
        idx = 0
        while True:
            if raw_string[idx] != '#':
                break
            else:
                idx += 1       
        first_line = f"{'#'*idx} System: {context_name} - "          
        return first_line + raw_string[idx:]


    def build_info_dict(self)-> OrderedDict:
        
        """Returns a dictionary of the markdown-formatted documentation of `self.system`
        and its children.

        Returns
        -------
        OrderedDict
            Ordered dictionary with key is contextual name of system and value is its
            documentation.
        """        

        ret = OrderedDict()
        root_name = self.system.name
        def _build_info(sys: System, ret):
            if sys.parent is None:
                context_name = root_name
            else:
                context_name = f"{root_name}.{sys.contextual_name}"
            ret[context_name] = SystemInfoComponent.create_html(sys, context_name)
            for child in sys.children.values():
                _build_info(child, ret)

        _build_info(self.system, ret)
        return ret

    def _handle_button_msg(self, model: Any, content: Dict, buffers: List)-> None:
        
        """Handlers for message of System Info Widget from front end.
        
        Parameters
        ----------
        model: Any
            Current System Info Widget model (unused)

        content: Dict
            Content of message.
        
        """        
        
        if content["action"] == f"{self.name}::getData":
            html_dict = self.build_info_dict()
            self.send(
                {
                    "type": f"{self.name}::infoData",
                    "payload": html_dict,
                }
            )

    def computed_notification(self) -> None:

        """Callback to be called after `computed` signal of `self.system`.
                
        """         
        html_dict = self.build_info_dict()
        self.send(
            {
                "type": f"{self.name}::updateData",
                "payload": html_dict,
            }
        )        



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

"""
This file originates from the 'jupyter-packaging' package, and
contains a set of useful utilities for including npm packages
within a Python package.
"""
from collections import defaultdict
from os.path import join as pjoin
from pathlib import Path
import io
import os
import functools
import pipes
import re
import shlex
import subprocess
import sys


# BEFORE importing distutils, remove MANIFEST. distutils doesn't properly
# update it when the contents of directories change.
if os.path.exists('MANIFEST'): os.remove('MANIFEST')


from setuptools import Command
from setuptools.command.build_py import build_py

# Note: distutils must be imported after setuptools
from distutils import log

from setuptools.command.sdist import sdist
from setuptools.command.develop import develop
from setuptools.command.bdist_egg import bdist_egg

try:
    from wheel.bdist_wheel import bdist_wheel
except ImportError:
    bdist_wheel = None

if sys.platform == 'win32':
    from subprocess import list2cmdline
else:
    def list2cmdline(cmd_list):
        return ' '.join(map(pipes.quote, cmd_list))


__version__ = '0.7.12'

# ---------------------------------------------------------------------------
# Top Level Variables
# ---------------------------------------------------------------------------

SEPARATORS = os.sep if os.altsep is None else os.sep + os.altsep

if "--skip-npm" in sys.argv:
    print("Skipping npm install as requested.")
    skip_npm = True
    sys.argv.remove("--skip-npm")
else:
    skip_npm = False


# ---------------------------------------------------------------------------
# Public Functions
# ---------------------------------------------------------------------------

def get_version(file, name='__version__'):
    """Get the version of the package from the given file by
    executing it and extracting the given `name`.
    """
    path = os.path.realpath(file)
    version_ns = {}
    with io.open(path, encoding="utf8") as f:
        exec(f.read(), {}, version_ns)
    return version_ns[name]


def ensure_python(specs):
    """Given a list of range specifiers for python, ensure compatibility.
    """
    import warnings
    warnings.warn(
        'Deprecated, please use python_requires in the setup function directly',
        category=DeprecationWarning
    )
    if not isinstance(specs, (list, tuple)):
        specs = [specs]
    v = sys.version_info
    part = '%s.%s' % (v.major, v.minor)
    for spec in specs:
        if part == spec:
            return
        try:
            if eval(part + spec):
                return
        except SyntaxError:
            pass
    raise ValueError('Python version %s unsupported' % part)


def find_packages(top):
    """
    Find all of the packages.
    """
    import warnings
    warnings.warn(
        'Deprecated, please use setuptools.find_packages',
        category=DeprecationWarning
    )
    from setuptools import find_packages as fp
    return fp(top)


def update_package_data(distribution):
    """update build_py options to get package_data changes"""
    build_py = distribution.get_command_obj('build_py')
    build_py.finalize_options()


class bdist_egg_disabled(bdist_egg):
    """Disabled version of bdist_egg

    Prevents setup.py install performing setuptools' default easy_install,
    which it should never ever do.
    """
    def run(self):
        sys.exit("Aborting implicit building of eggs. Use `pip install .` "
                 " to install from source.")


def create_cmdclass(prerelease_cmd=None, package_data_spec=None,
                    data_files_spec=None, exclude=None):
    """Create a command class with the given optional prerelease class.

    Parameters
    ----------
    prerelease_cmd: (name, Command) tuple, optional
        The command to run before releasing.
    package_data_spec: dict, optional
        A dictionary whose keys are the dotted package names and
        whose values are a list of glob patterns.
    data_files_spec: list, optional
        A list of (path, dname, pattern) tuples where the path is the
        `data_files` install path, dname is the source directory, and the
        pattern is a glob pattern.
    exclude: function
        A function which takes a string filename and returns True if the
        file should be excluded from package data and data files, False otherwise.

    Notes
    -----
    We use specs so that we can find the files *after* the build
    command has run.

    The package data glob patterns should be relative paths from the package
    folder containing the __init__.py file, which is given as the package
    name.
    e.g. `dict(foo=['bar/*', 'baz/**'])`

    The data files directories should be absolute paths or relative paths
    from the root directory of the repository.  Data files are specified
    differently from `package_data` because we need a separate path entry
    for each nested folder in `data_files`, and this makes it easier to
    parse.
    e.g. `('share/foo/bar', 'pkgname/bizz, '*')`
    """
    wrapped = [prerelease_cmd] if prerelease_cmd else []
    if package_data_spec or data_files_spec:
        wrapped.append('handle_files')

    wrapper = functools.partial(_wrap_command, wrapped)
    handle_files = _get_file_handler(package_data_spec, data_files_spec, exclude)
    develop_handler = _get_develop_handler()

    if 'bdist_egg' in sys.argv:
        egg = wrapper(bdist_egg, strict=True)
    else:
        egg = bdist_egg_disabled

    is_repo = os.path.exists('.git')

    cmdclass = dict(
        build_py=wrapper(build_py, strict=is_repo),
        bdist_egg=egg,
        sdist=wrapper(sdist, strict=True),
        handle_files=handle_files,
    )

    if bdist_wheel:
        cmdclass['bdist_wheel'] = wrapper(bdist_wheel, strict=True)

    cmdclass['develop'] = wrapper(develop_handler, strict=True)
    return cmdclass


def command_for_func(func):
    """Create a command that calls the given function."""

    class FuncCommand(BaseCommand):

        def run(self):
            func()
            update_package_data(self.distribution)

    return FuncCommand


def run(cmd, **kwargs):
    """Echo a command before running it."""
    log.info('> ' + list2cmdline(cmd))
    kwargs.setdefault('shell', os.name == 'nt')
    if not isinstance(cmd, (list, tuple)) and os.name != 'nt':
        cmd = shlex.split(cmd)
    cmd_path = which(cmd[0])
    if not cmd_path:
        sys.exit("Aborting. Could not find cmd (%s) in path. "
                 "If command is not expected to be in user's path, "
                 "use an absolute path." % cmd[0])
    cmd[0] = cmd_path
    return subprocess.check_call(cmd, **kwargs)


def is_stale(target, source):
    """Test whether the target file/directory is stale based on the source
       file/directory.
    """
    if not os.path.exists(target):
        return True
    target_mtime = recursive_mtime(target) or 0
    return compare_recursive_mtime(source, cutoff=target_mtime)


class BaseCommand(Command):
    """Empty command because Command needs subclasses to override too much"""
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def get_inputs(self):
        return []

    def get_outputs(self):
        return []


def combine_commands(*commands):
    """Return a Command that combines several commands."""

    class CombinedCommand(Command):
        user_options = []

        def initialize_options(self):
            self.commands = []
            for C in commands:
                self.commands.append(C(self.distribution))
            for c in self.commands:
                c.initialize_options()

        def finalize_options(self):
            for c in self.commands:
                c.finalize_options()

        def run(self):
            for c in self.commands:
                c.run()
    return CombinedCommand


def compare_recursive_mtime(path, cutoff, newest=True):
    """Compare the newest/oldest mtime for all files in a directory.

    Cutoff should be another mtime to be compared against. If an mtime that is
    newer/older than the cutoff is found it will return True.
    E.g. if newest=True, and a file in path is newer than the cutoff, it will
    return True.
    """
    if os.path.isfile(path):
        mt = mtime(path)
        if newest:
            if mt > cutoff:
                return True
        elif mt < cutoff:
            return True
    for dirname, _, filenames in os.walk(path, topdown=False):
        for filename in filenames:
            mt = mtime(pjoin(dirname, filename))
            if newest:  # Put outside of loop?
                if mt > cutoff:
                    return True
            elif mt < cutoff:
                return True
    return False


def recursive_mtime(path, newest=True):
    """Gets the newest/oldest mtime for all files in a directory."""
    if os.path.isfile(path):
        return mtime(path)
    current_extreme = None
    for dirname, dirnames, filenames in os.walk(path, topdown=False):
        for filename in filenames:
            mt = mtime(pjoin(dirname, filename))
            if newest:  # Put outside of loop?
                if mt >= (current_extreme or mt):
                    current_extreme = mt
            elif mt <= (current_extreme or mt):
                current_extreme = mt
    return current_extreme


def mtime(path):
    """shorthand for mtime"""
    return os.stat(path).st_mtime


def install_npm(path=None, build_dir=None, source_dir=None, build_cmd='build',
                force=False, npm=None):
    """Return a Command for managing an npm installation.

    Note: The command is skipped if the `--skip-npm` flag is used.

    Parameters
    ----------
    path: str, optional
        The base path of the node package. Defaults to the current directory.
    build_dir: str, optional
        The target build directory.  If this and source_dir are given,
        the JavaScript will only be build if necessary.
    source_dir: str, optional
        The source code directory.
    build_cmd: str, optional
        The npm command to build assets to the build_dir.
    npm: str or list, optional.
        The npm executable name, or a tuple of ['node', executable].
    """

    class NPM(BaseCommand):
        description = 'install package.json dependencies using npm'

        def run(self):
            if skip_npm:
                log.info('Skipping npm-installation')
                return
            node_package = path or os.path.abspath(os.getcwd())
            node_modules = pjoin(node_package, 'node_modules')
            is_yarn = os.path.exists(pjoin(node_package, 'yarn.lock'))

            npm_cmd = npm

            if npm is None:
                if is_yarn:
                    npm_cmd = ['yarn']
                else:
                    npm_cmd = ['npm']

            if not which(npm_cmd[0]):
                log.error("`{0}` unavailable.  If you're running this command "
                          "using sudo, make sure `{0}` is available to sudo"
                          .format(npm_cmd[0]))
                return

            if build_dir and source_dir and not force:
                should_build = is_stale(build_dir, source_dir)
            else:
                should_build = True

            if should_build:
                log.info('Installing build dependencies with npm.  This may '
                         'take a while...')
                run(npm_cmd + ['install'], cwd=node_package)
                if build_cmd:
                    run(npm_cmd + ['run', build_cmd], cwd=node_package)

    return NPM


def ensure_targets(targets):
    """Return a Command that checks that certain files exist.

    Raises a ValueError if any of the files are missing.

    Note: The check is skipped if the `--skip-npm` flag is used.
    """

    class TargetsCheck(BaseCommand):
        def run(self):
            if skip_npm:
                log.info('Skipping target checks')
                return
            missing = [t for t in targets if not os.path.exists(t)]
            if missing:
                raise ValueError(('missing files: %s' % missing))

    return TargetsCheck


def skip_if_exists(paths, CommandClass):
    """Skip a command if list of paths exists."""
    def should_skip():
        return all(Path(path).exists() for path in paths)
    class SkipIfExistCommand(Command):
        def initialize_options(self):
            if not should_skip():
                self.command = CommandClass(self.distribution)
                self.command.initialize_options()
            else:
                self.command = None

        def finalize_options(self):
            if self.command is not None:
                self.command.finalize_options()

        def run(self):
            if self.command is not None:
                self.command.run()

    return SkipIfExistCommand


# `shutils.which` function copied verbatim from the Python-3.3 source.
def which(cmd, mode=os.F_OK | os.X_OK, path=None):
    """Given a command, mode, and a PATH string, return the path which
    conforms to the given mode on the PATH, or None if there is no such
    file.
    `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
    of os.environ.get("PATH"), or can be overridden with a custom search
    path.
    """

    # Check that a given file can be accessed with the correct mode.
    # Additionally check that `file` is not a directory, as on Windows
    # directories pass the os.access check.
    def _access_check(fn, mode):
        return (os.path.exists(fn) and os.access(fn, mode) and
                not os.path.isdir(fn))

    # Short circuit. If we're given a full path which matches the mode
    # and it exists, we're done here.
    if _access_check(cmd, mode):
        return cmd

    path = (path or os.environ.get("PATH", os.defpath)).split(os.pathsep)

    if sys.platform == "win32":
        # The current directory takes precedence on Windows.
        if os.curdir not in path:
            path.insert(0, os.curdir)

        # PATHEXT is necessary to check on Windows.
        pathext = os.environ.get("PATHEXT", "").split(os.pathsep)
        # See if the given file matches any of the expected path extensions.
        # This will allow us to short circuit when given "python.exe".
        matches = [cmd for ext in pathext if cmd.lower().endswith(ext.lower())]
        # If it does match, only test that one, otherwise we have to try
        # others.
        files = [cmd] if matches else [cmd + ext.lower() for ext in pathext]
    else:
        # On other platforms you don't have things like PATHEXT to tell you
        # what file suffixes are executable, so just pass on cmd as-is.
        files = [cmd]

    seen = set()
    for dir in path:
        dir = os.path.normcase(dir)
        if dir not in seen:
            seen.add(dir)
            for thefile in files:
                name = os.path.join(dir, thefile)
                if _access_check(name, mode):
                    return name
    return None


# ---------------------------------------------------------------------------
# Private Functions
# ---------------------------------------------------------------------------


def _wrap_command(cmds, cls, strict=True):
    """Wrap a setup command

    Parameters
    ----------
    cmds: list(str)
        The names of the other commands to run prior to the command.
    strict: boolean, optional
        Whether to raise errors when a pre-command fails.
    """
    class WrappedCommand(cls):

        def run(self):
            if not getattr(self, 'uninstall', None):
                try:
                    [self.run_command(cmd) for cmd in cmds]
                except Exception:
                    if strict:
                        raise
                    else:
                        pass
            # update package data
            update_package_data(self.distribution)

            result = cls.run(self)
            return result
    return WrappedCommand


def _get_file_handler(package_data_spec, data_files_spec, exclude=None):
    """Get a package_data and data_files handler command.
    """
    class FileHandler(BaseCommand):

        def run(self):
            package_data = self.distribution.package_data
            package_spec = package_data_spec or dict()

            for (key, patterns) in package_spec.items():
                files =  _get_package_data(key, patterns)
                if exclude is not None:
                    files = [f for f in files if not exclude(f)]
                package_data[key] =files

            self.distribution.data_files = _get_data_files(
                data_files_spec, self.distribution.data_files, exclude=exclude
            )

    return FileHandler


def _get_develop_handler():
    """Get a handler for the develop command"""
    class _develop(develop):

        def install_for_development(self):
            self.finalize_options()
            super(_develop, self).install_for_development()
            self.run_command('handle_files')
            prefix = self.install_base or self.prefix or sys.prefix
            for target_dir, filepaths in self.distribution.data_files:
                for filepath in filepaths:
                    filename = os.path.basename(filepath)
                    target = os.path.join(prefix, target_dir, filename)
                    self.mkpath(os.path.dirname(target))
                    outf, copied = self.copy_file(filepath, target)

    return _develop


def _glob_pjoin(*parts):
    """Join paths for glob processing"""
    if parts[0] in ('.', ''):
        parts = parts[1:]
    return pjoin(*parts).replace(os.sep, '/')


def _get_data_files(data_specs, existing, *, top=None, exclude=None):
    """Expand data file specs into valid data files metadata.

    Parameters
    ----------
    data_specs: list of tuples
        See [create_cmdclass] for description.
    existing: list of tuples
        The existing distribution data_files metadata.
    top: str, optional
        The top directory
    exclude: func, optional
        Function used to test whether to exclude a file

    Returns
    -------
    A valid list of data_files items.
    """
    if top is None:
        top = os.path.abspath(os.getcwd())
    # Extract the existing data files into a staging object.
    file_data = defaultdict(list)
    for (path, files) in existing or []:
        file_data[path] = files

    # Extract the files and assign them to the proper data
    # files path.
    for (path, dname, pattern) in data_specs or []:
        if os.path.isabs(dname):
            dname = os.path.relpath(dname, top)

        dname = dname.replace(os.sep, '/').rstrip('/')
        offset = 0 if dname in ('.', '') else len(dname) + 1
        files = _get_files(_glob_pjoin(dname, pattern), top=top)

        for fname in files:
            # Normalize the path.
            root = os.path.dirname(fname)
            full_path = _glob_pjoin(path, root[offset:])
            if full_path.endswith('/'):
                full_path = full_path[:-1]
            if exclude is not None and exclude(fname):
                continue
            file_data[full_path].append(fname)

    # Construct the data files spec.
    data_files = []
    for (path, files) in file_data.items():
        data_files.append((path, files))
    return data_files


def _get_files(file_patterns, top=None):
    """Expand file patterns to a list of paths.

    Parameters
    -----------
    file_patterns: list or str
        A list of glob patterns for the data file locations.
        The globs can be recursive if they include a `**`.
        They should be relative paths from the top directory or
        absolute paths.
    top: str
        the directory to consider for data files

    Note:
    Files in `node_modules` are ignored.
    """
    if top is None:
        top = os.path.abspath(os.getcwd())
    if not isinstance(file_patterns, (list, tuple)):
        file_patterns = [file_patterns]

    for i, p in enumerate(file_patterns):
        if os.path.isabs(p):
            file_patterns[i] = os.path.relpath(p, top)

    matchers = [_compile_pattern(p) for p in file_patterns]

    files = set()

    for root, dirnames, filenames in os.walk(top):
        # Don't recurse into node_modules
        if 'node_modules' in dirnames:
            dirnames.remove('node_modules')
        for m in matchers:
            for filename in filenames:
                fn = os.path.relpath(_glob_pjoin(root, filename), top)
                fn = fn.replace(os.sep, '/')
                if m(fn):
                    files.add(fn)

    return list(files)


def _get_package_data(root, file_patterns=None):
    """Expand file patterns to a list of `package_data` paths.

    Parameters
    -----------
    root: str
        The relative path to the package root from the current dir.
    file_patterns: list or str, optional
        A list of glob patterns for the data file locations.
        The globs can be recursive if they include a `**`.
        They should be relative paths from the root or
        absolute paths.  If not given, all files will be used.

    Note:
    Files in `node_modules` are ignored.
    """
    if file_patterns is None:
        file_patterns = ['*']
    return _get_files(file_patterns, _glob_pjoin(os.path.abspath(os.getcwd()), root))


def _compile_pattern(pat, ignore_case=True):
    """Translate and compile a glob pattern to a regular expression matcher."""
    if isinstance(pat, bytes):
        pat_str = pat.decode('ISO-8859-1')
        res_str = _translate_glob(pat_str)
        res = res_str.encode('ISO-8859-1')
    else:
        res = _translate_glob(pat)
    flags = re.IGNORECASE if ignore_case else 0
    return re.compile(res, flags=flags).match


def _iexplode_path(path):
    """Iterate over all the parts of a path.

    Splits path recursively with os.path.split().
    """
    (head, tail) = os.path.split(path)
    if not head or (not tail and head == path):
        if head:
            yield head
        if tail or not head:
            yield tail
        return
    for p in _iexplode_path(head):
        yield p
    yield tail


def _translate_glob(pat):
    """Translate a glob PATTERN to a regular expression."""
    translated_parts = []
    for part in _iexplode_path(pat):
        translated_parts.append(_translate_glob_part(part))
    os_sep_class = '[%s]' % re.escape(SEPARATORS)
    res = _join_translated(translated_parts, os_sep_class)
    return '(?ms){res}\\Z'.format(res=res)


def _join_translated(translated_parts, os_sep_class):
    """Join translated glob pattern parts.

    This is different from a simple join, as care need to be taken
    to allow ** to match ZERO or more directories.
    """
    res = ''
    for part in translated_parts[:-1]:
        if part == '.*':
            # drop separator, since it is optional
            # (** matches ZERO or more dirs)
            res += part
        else:
            res += part + os_sep_class

    if translated_parts[-1] == '.*':
        # Final part is **
        res += '.+'
        # Follow stdlib/git convention of matching all sub files/directories:
        res += '({os_sep_class}?.*)?'.format(os_sep_class=os_sep_class)
    else:
        res += translated_parts[-1]
    return res


def _translate_glob_part(pat):
    """Translate a glob PATTERN PART to a regular expression."""
    # Code modified from Python 3 standard lib fnmatch:
    if pat == '**':
        return '.*'
    i, n = 0, len(pat)
    res = []
    while i < n:
        c = pat[i]
        i = i + 1
        if c == '*':
            # Match anything but path separators:
            res.append('[^%s]*' % SEPARATORS)
        elif c == '?':
            res.append('[^%s]?' % SEPARATORS)
        elif c == '[':
            j = i
            if j < n and pat[j] == '!':
                j = j + 1
            if j < n and pat[j] == ']':
                j = j + 1
            while j < n and pat[j] != ']':
                j = j + 1
            if j >= n:
                res.append('\\[')
            else:
                stuff = pat[i:j].replace('\\', '\\\\')
                i = j + 1
                if stuff[0] == '!':
                    stuff = '^' + stuff[1:]
                elif stuff[0] == '^':
                    stuff = '\\' + stuff
                res.append('[%s]' % stuff)
        else:
            res.append(re.escape(c))
    return ''.join(res)


from copy import deepcopy
from enum import Enum
import logging
from math import inf
from typing import Any, Optional, Tuple
from weakref import ref

from cosapp.ports import Validity
from cosapp.systems import System
from ipywidgets import (
    BoundedFloatText,
    VBox,
    Button,
    Dropdown,
    FloatText,
    HBox,
    HTML,
    Label,
    Layout,
    Text,
    Widget,
)

from cosapp_lab.widgets.legacy.widgetlogger import WidgetLogger
from cosapp_lab.widgets.legacy.utils import default_step

logger = logging.getLogger(__name__)


class SystemWidget(VBox):
    """Abstract base class for Jupyter Widget used with CoSApp System.

    Class Attributes
    ----------------
    auto_run : bool
        Should the CoSApp `System` be executed at each value change.
    default_layout : Dict[str, str]
        Default box layout

    Parameters
    ----------
    system : System
        System in which belong the variable
    name : str
        Full variable name of the variable in the CoSApp System
    alias : str, optional
        Alias for the variable name to be shown in the UI; default None (i.e. alias = name)
    children : Tuple[Widget], optional
        Tuple of children widget to add to this box; default empty tuple
    kwargs
        Additional keyword arguments are passed to the Box layout
    """

    auto_run = True
    default_layout = {"flex_flow": "column nowrap"}

    def __init__(
        self,
        system: System,
        name: str,
        alias: Optional[str] = None,
        children: Tuple[Widget] = (),
        **kwargs
    ):
        self._name = name
        self._system = ref(system)
        self._init_value = deepcopy(system[name])
        if alias is None:
            alias = name
        layout = kwargs
        layout.update(SystemWidget.default_layout)
        self._icon = HTML("", layout=Layout(flex="0 0 auto"))
        label = Label(alias, layout=Layout(flex="1 1 auto"))
        reset_but = Button(
            icon="undo",
            tooltip="Reset to initial value",
            layout=Layout(flex="0 0 auto", width="auto"),
        )
        reset_but.on_click(self.reset)
        super(SystemWidget, self).__init__(
            children=(HBox([self._icon, label, reset_but]), *children),
            layout=Layout(**layout),
        )

        # Connect events
        system.setup_ran.connect(self.update)

    @staticmethod
    def get_from_name(
        system: System, name: str, alias: Optional[str] = None, **kwargs
    ) -> "SystemWidget":
        """Create a handler widget for a variable specified by its name.

        Parameters
        ----------
        system : System
            System in which belong the variable
        name : str
            Full variable name of the variable in the CoSApp System
        alias : str, optional
            Alias for the variable name to be shown in the UI; default None (i.e. alias = name)
        kwargs
            Additional keyword arguments are passed to the widget

        Returns
        -------
        SystemWidget
            The widget to handle the variable.
        """

        if name not in system:
            raise ValueError(
                "Variable '{}' does not exist in System '{}'.".format(name, system.name)
            )

        value = system[name]

        if isinstance(value, Enum):
            return EnumData(system, name, alias=alias, **kwargs)
        elif isinstance(value, (str, bytes)):
            return StringData(system, name, alias=alias, **kwargs)

        try:
            float(value)
        except (ValueError, TypeError):
            pass
        else:
            return NumberData(system, name, alias=alias, **kwargs)

        try:
            float(list(value)[0])
        except (ValueError, TypeError):
            pass
        else:
            return ArrayData(system, name, alias=alias, **kwargs)

        raise NotImplementedError(
            "Variable '{}' has not a supported type for displaying in a widget.".format(
                name
            )
        )

    def on_value_changed(self, value: Any):
        """Callback following a widget value change."""
        s = self._system()
        if s is None:
            return

        s[self._name] = value
        # Valid the new value
        valid = s.check(self._name)
        if valid == Validity.OK:
            self._icon.value = ""
        elif valid == Validity.WARNING:
            self._icon.value = (
                "<i class='fa fa-bell' style='color:var(--jp-warn-color1);'></i>"
            )
        elif valid == Validity.ERROR:
            self._icon.value = "<i class='fa fa-exclamation-triangle' style='color:var(--jp-error-color1);'></i>"

        if SystemWidget.auto_run:
            try:
                s.run_drivers()
            except Exception as e:
                logger.exception(e)

    def reset(self, event=None):
        """Reset the current widget value with the initial value."""
        raise NotImplementedError(
            "`SystemWidget.reset` is an abstract method that must be implemented in subclasses."
        )

    def update(self, event=None):
        """Update the current widget value from the CoSApp System."""
        raise NotImplementedError(
            "`SystemWidget.update` is an abstract method that must be implemented in subclasses."
        )


class UniqueData(SystemWidget):
    """Widget to update a unique value.

    Parameters
    ----------
    system : System
        System in which belong the variable
    name : str
        Full variable name of the variable in the CoSApp System
    child : Widget
        The widget modifying the value
    alias : str, optional
        Alias for the variable name to be shown in the UI; default None (i.e. alias = name)
    """

    def __init__(
        self,
        system: System,
        name: str,
        child: Widget,
        alias: Optional[str] = None,
        **kwargs
    ):
        self.widget = child
        super(UniqueData, self).__init__(
            system, name, alias=alias, children=(child,), **kwargs
        )
        self.widget.observe(self.on_value, names="value")

    def on_value(self, change):
        """Callback method when widget value changes."""
        self.on_value_changed(change["new"])

    def _set_value(self, value):
        run_status = SystemWidget.auto_run
        SystemWidget.auto_run = False
        self.widget.value = value
        SystemWidget.auto_run = run_status

    def reset(self, event=None):
        self._set_value(self._init_value)

    def update(self, event=None):
        s = self._system()
        if s is not None:
            self._set_value(s[self._name])


class NumberData(UniqueData):
    """Widget to update a number.

    Parameters
    ----------
    system : System
        System in which belong the variable
    name : str
        Full variable name of the variable in the CoSApp System
    alias : str, optional
        Alias for the variable name to be shown in the UI; default None (i.e. alias = name)
    min : number, optional
        Minimum value for the variable; default None
    max : number, optional
        Maximum value for the variable; default None
    step : number, optional
        Increment of variable; default None
    """

    def __init__(
        self, system: System, name: str, alias: Optional[str] = None, **kwargs
    ):
        value = float(system[name])
        # TODO use validation range to potentially set min and max
        min = kwargs.pop("min", None)
        max = kwargs.pop("max", None)
        step = kwargs.pop("step", None)  # TODO estimate step from value

        if min is None and max is None:
            widget = FloatText(
                value=value,
                step=step or default_step(value),
                layout=Layout(width="auto"),
            )
        else:
            if not (min is None or max is None) and step is None:
                step = 0.05 * (max - min)  # Take step = 5% range
            min = min if min is not None else -1.0 * inf
            max = max if max is not None else inf
            widget = BoundedFloatText(
                value=value,
                max=max,
                min=min,
                step=step or default_step(value),
                layout=Layout(width="auto"),
            )

        super(NumberData, self).__init__(system, name, widget, alias=alias, **kwargs)


class EnumData(UniqueData):
    """Widget to update a enum.

    Parameters
    ----------
    system : System
        System in which belong the variable
    name : str
        Full variable name of the variable in the CoSApp System
    alias : str, optional
        Alias for the variable name to be shown in the UI; default None (i.e. alias = name)
    """

    def __init__(
        self, system: System, name: str, alias: Optional[str] = None, **kwargs
    ):
        value = system[name]
        widget = Dropdown(
            value=value,
            options=type(value)._value2member_map_,
            layout=Layout(width="auto"),
        )

        super(EnumData, self).__init__(system, name, widget, alias=alias, **kwargs)


class StringData(UniqueData):
    """Widget to update a string.

    Parameters
    ----------
    system : System
        System in which belong the variable
    name : str
        Full variable name of the variable in the CoSApp System
    alias : str, optional
        Alias for the variable name to be shown in the UI; default None (i.e. alias = name)
    options : str, optional
        List of valid strings; default None (i.e. user is free to type the data)
    """

    def __init__(
        self, system: System, name: str, alias: Optional[str] = None, **kwargs
    ):
        value = system[name]
        options = kwargs.pop("options", None)
        if options is None:
            widget = Text(value=value, layout=Layout(width="auto"))
        else:
            options = set(options)
            if value not in options:
                options.add(value)
            widget = Dropdown(value=value, options=options, layout=Layout(width="auto"))

        super(StringData, self).__init__(system, name, widget, alias=alias, **kwargs)

    def reset(self, event=None):
        if (
            isinstance(self.widget, Dropdown)
            and self._init_value not in self.widget.options
        ):
            options = set(self.widget.options)
            self.widget.options = options.add(self._init_value)

        super(StringData, self).reset(event)

    def update(self, event=None):
        s = self._system()
        if s is not None:
            v = s[self._name]
            if isinstance(self.widget, Dropdown) and v not in self.widget.options:
                options = set(self.widget.options)
                self.widget.options = options.add(v)

        super(StringData, self).update(event)


class IndexFloatText(FloatText):
    """FloatText widget extended to keep track of the index of the float in the associated array.

    Parameters
    ----------
    index : int
        Index of the number in the container array
    """

    def __init__(self, index: int, **kwargs):
        super(IndexFloatText, self).__init__(**kwargs)
        self.index = index


class ArrayData(SystemWidget):
    """Widget to update an iterable of number.

    Parameters
    ----------
    system : System
        System in which belong the variable
    name : str
        Full variable name of the variable in the CoSApp System
    alias : str, optional
        Alias for the variable name to be shown in the UI; default None (i.e. alias = name)
    """

    def __init__(
        self, system: System, name: str, alias: Optional[str] = None, **kwargs
    ):
        self.value = deepcopy(system[name])
        self.vector = list()

        for i, vi in enumerate(self.value):
            f = IndexFloatText(i, value=vi, layout=Layout(flex="1 1 80px"))
            f.observe(self.on_value, names="value")
            self.vector.append(f)

        super(ArrayData, self).__init__(
            system, name, alias=alias, children=(HBox(children=self.vector),), **kwargs
        )

    def on_value(self, changes):
        widget = changes["owner"]
        self.value[widget.index] = changes["new"]
        self.on_value_changed(self.value)

    def _set_value(self, value):
        self.value = deepcopy(value)
        run_status = SystemWidget.auto_run
        SystemWidget.auto_run = False

        for i, widget in enumerate(self.vector):
            widget.value = self.value[i]
        SystemWidget.auto_run = run_status

    def reset(self, event=None):
        """Reset the current widget value with the initial value."""
        self._set_value(self._init_value)

    def update(self, event=None):
        """Update the current widget value from the CoSApp System."""
        s = self._system()
        if s is not None:
            self._set_value(s[self._name])



from jupyter_client.blocking import BlockingKernelClient
import json
import queue
from typing import Tuple, Union, Dict


class CosappKernelConnetion:
    """
    A class to encapsulate a `BlockingKernelClient` of jupyter.

    Parameters
    ----------
    connection_file: str
        Path to the `connection_file` of the kernel we want to connect with.

    system_name: str
        The name of Cosapp system, this name is used to get correct `SysExplorer` instance.

    Attributes
    ----------
    client : BlockingKernelClient
        The blocking client instance of connected kernel

    idle_count : int
        Counter used to get output of execution, this counter is reset
        to 0 each time a command is ternimated.

    sysexplorer : str
        Name of the `SysExplorer` instance associated with `system_name`
        in connected kernel.

    system_name : str
        Name of the `cosapp` system instance in connected kernel.

    """

    def __init__(self, connection_file: str, system_name: str):
        self.client = BlockingKernelClient()

        self.client.load_connection_file(connection_file)
        self.client.start_channels()
        self.idle_count = 0
        self.sysexplorer, self.system_name = self.get_variable_name(system_name)
        if self.sysexplorer == None:
            raise NameError("Requested SysExplorer instance can not be found")

    def execute(self, cmd: str, to_json=True) -> Tuple[str, Union[str, Dict]]:
        """
        Execute a command in connected kernel and return the output

        Parameters
        ----------
        cmd: str
            Python command to be executed.

        to_json: bool
            Flag to convert output string into a Dict.

        Returns
        -------

        value_type : str
            Type of return value.

        value : Union[str, Dict]
            Depend on `to_json` flag, `value` will be a string or a dict .
        """
        msg_id = self.client.execute(cmd)
        value = None
        value_type = "unknown"
        while True:
            try:
                io_msg = self.client.get_iopub_msg(timeout=30)["content"]
                if "data" in io_msg:
                    if "text/plain" in io_msg["data"]:
                        value_string: str = io_msg["data"]["text/plain"]
                        value_type = "data"
                        if to_json:
                            if value_string.startswith("'"):
                                value = json.loads(value_string[1:-1].replace("'", '"'))
                            elif value_string.startswith("{"):
                                value = json.loads(value_string.replace("'", '"'))
                            else:
                                value = ["Json parser error", "Data format incorrect"]
                                value_type = "error"
                        else:
                            value = value_string

                elif "name" in io_msg:
                    value = io_msg["text"]
                    value_type = "text"
                elif "traceback" in io_msg:
                    value = [io_msg["ename"], io_msg["evalue"]]
                    value_type = "error"
                elif (
                    "execution_state" in io_msg and io_msg["execution_state"] == "idle"
                ):
                    if self.idle_count == 0:
                        if value is None and value_type == "unknown":
                            self.idle_count += 1
                        else:
                            break
                    else:
                        break

            except queue.Empty as e:
                if value is None and value_type == "unknown":
                    return "error", ["error log", str(e)]
                else:
                    print(e)
        return value_type, value

    def get_variable_name(self, requested_system_name: str) -> Tuple[str]:
        """
        Get the `SysExplorer` instance associated with `requested_system_name`
        in connected kernel.

        Parameters
        ----------
        requested_system_name: str
            Name of the `cosapp` system instance in connected kernel.
        """
        ret = None
        system_name = None
        flag, var_string = self.execute("%who", False)
        var_list = var_string.replace("\n", "").split("\t")
        if len(var_list) == 1 and var_list[0] == "Interactive namespace is empty.":
            return None, None
        for var in var_list:
            var_name = var.strip()
            if len(var_name) > 0:
                flag, var_type = self.execute(
                    f"isinstance({var_name}, SysExplorer)", False
                )
                if flag == "data" and var_type == "True":
                    flag, system_name_test = self.execute(
                        f"{var_name}._system().name", False
                    )
                    if (
                        system_name_test.replace("'", "").replace('"', "")
                        == requested_system_name
                    ):
                        ret = var_name
                        system_name = system_name_test
        return ret, system_name

    def disconnect(self):
        """
        Stop the connection to kernel
        """
        self.client.stop_channels()



import json
from typing import Dict, NoReturn
from notebook.base.handlers import APIHandler
from notebook.notebookapp import NotebookWebApplication
from notebook.utils import url_path_join
from pathlib import Path
import shutil
from .cosapp_kernel import CosappKernelConnetion

COSAPP_URL_START = "cosapp/server/start"
COSAPP_URL_STOP = "cosapp/server/stop"
COSAPP_URL_INFO = "cosapp/server/info"
COSAPP_URL_RUN = "cosapp/server/run"
COSAPP_CONFIG_DIR = Path.home() / ".cosapp.d"


def setup_cosapp_server_handlers(web_app: NotebookWebApplication):
    """Setup handlers for to cosapp server.

    Parameters
    ----------
    web_app : notebook.notebookapp.NotebookWebApplication
        The notebook web application
    """
    host_pattern = ".*$"

    def build_url(url_extension: str) -> str:
        return url_path_join(web_app.settings["base_url"], url_extension)

    web_app.add_handlers(
        host_pattern,
        [
            (build_url(COSAPP_URL_START), StartSessionHandler),
            (build_url(COSAPP_URL_STOP), StopSessionHandler),
            (build_url(COSAPP_URL_RUN), RunSessionHandler),
            (build_url(COSAPP_URL_INFO), InfoSessionHandler),
        ],
    )


class CustomAPI(APIHandler):
    """
    Custom `APIHandler` to bypass `xsrf` cookie check and
    close kernel connection on finish.
    """

    def check_xsrf_cookie(self):
        return

    def on_finish(self):
        if hasattr(self, "kc") and self.kc is not None:
            self.kc.disconnect()

    def create_kernel_connection(self, token: str) -> CosappKernelConnetion:
        """
        Helper function to connect to kernel by using
        configuration stored in `server` folder of `COSAPP_CONFIG_DIR`
        """
        folder_path = COSAPP_CONFIG_DIR / "server" / token
        if folder_path.exists():
            try:
                kernel_config = folder_path / "config.json"
                with open(kernel_config, "r") as f:
                    config = json.load(f)
                connection_file = config["connection"]
                request_name = config["system_name"]
                kc = CosappKernelConnetion(connection_file, request_name)
            except Exception as e:
                print(e)
                kc = None
        else:
            kc = None
        return kc


class StartSessionHandler(CustomAPI):
    """
    Handle start session request, this handler will clear
    all old configuration file in `COSAPP_CONFIG_DIR` before
    creating new session.
    """

    def get(self):
        self.finish("Start")

    def post(self):
        self.clear_server_folder()
        input_data = self.get_json_body()
        folder_path = COSAPP_CONFIG_DIR / "server" / input_data["token"]
        folder_path.mkdir(parents=True, exist_ok=True)
        kernel_config = folder_path / "config.json"
        with open(kernel_config, "w") as f:
            json.dump(input_data, f)

        self.finish("1")

    def clear_server_folder(self):
        folder_path = COSAPP_CONFIG_DIR / "server"
        for sub_folder in folder_path.glob("*"):
            config_path = folder_path / sub_folder
            try:
                kc = self.create_kernel_connection(sub_folder)
                if kc == None:
                    shutil.rmtree(config_path)
                else:
                    kc.disconnect()
            except:
                shutil.rmtree(config_path)


class StopSessionHandler(CustomAPI):
    """
    Handle stop session request, this handler will remove
    associated configuration file.
    """

    def get(self):
        self.finish("Stop")

    def post(self):
        input_data = self.get_json_body()
        folder_path = COSAPP_CONFIG_DIR / "server" / input_data["token"]
        if folder_path.exists():
            try:
                shutil.rmtree(folder_path)
                self.finish("1")
            except Exception as e:
                print(e)
                self.finish("-1")

        else:
            self.finish("-1")


class InfoSessionHandler(CustomAPI):
    """
    Handle system information request.
    """

    def post(self):

        input_data = self.get_json_body()
        server_msg = f"Received INFO request from {self.request.remote_ip}"
        self.kc = self.create_kernel_connection(input_data["token"])

        if self.kc is not None:
            self.kc.execute(f"{self.kc.sysexplorer}.update_server_log('{server_msg}')")
            msg = {}
            for name in ["children_list", "children_port", "children_drive"]:
                cmd = f"{self.kc.sysexplorer}.sys_data.{name}"
                try:
                    flag, ret = self.kc.execute(cmd)
                    msg[name] = ret
                except Exception as e:
                    print(e)
                    msg["error"] = e
                    break
            self.finish(msg)
        else:
            self.finish("-1")


class RunSessionHandler(CustomAPI):

    """
    Handle run system request.
    """

    def post(self):
        input_data = self.get_json_body()
        post_data: Dict = input_data["data"]
        requested_result = post_data["result"]
        server_msg = f"Received RUN request from {self.request.remote_ip}"
        self.kc = self.create_kernel_connection(input_data["token"])
        if self.kc is not None:
            self.kc.execute(f"{self.kc.sysexplorer}.update_server_log('{server_msg}')")
            param_dict = post_data["parameters"]
            post_content = {}
            for key, val in param_dict.items():
                post_content[f"{key}.{val[0]}.{val[1]}"] = val[2]
            content = json.dumps(
                {"action": "runSignal", "payload": post_content, "currentThread": "1"}
            )

            cmd = f"{self.kc.sysexplorer}._handle_button_msg('SysExplorerModel', {content}, [])"
            run_flag, run_log = self.kc.execute(cmd, False)
            if run_flag == "error":
                self.finish({"error": run_log, "result": None, "log": None})
            else:
                flag, exec_result = self.kc.execute(
                    f"{self.kc.sysexplorer}.sys_data.serialize_data_from_system()", True
                )
                ret = {}

                if len(requested_result) == 0:
                    ret = exec_result
                else:
                    for result_key in requested_result:
                        for key, value in exec_result.items():
                            if key.startswith(result_key):
                                ret[key] = value

                flag, exec_log = self.kc.execute(
                    f"{self.kc.sysexplorer}.server_log", False
                )
                self.finish({"error": None, "result": ret, "log": exec_log})
        else:
            self.finish("-1")



import os
import sys
from pathlib import Path
import logging


def get_abs_dir(dir_path: str, path: str) -> str:
    if os.path.isabs(path):
        return path
    else:
        return os.path.join(dir_path, path)


class DummyFile(object):
    def write(self, x):
        pass


class Suppressor(object):
    def __enter__(self):

        self.stdout = sys.stdout
        sys.stdout = self

    def __exit__(self, type, value, traceback):
        sys.stdout = self.stdout
        if type is not None:
            raise type

    def write(self, x):
        pass


def get_readme_from_path(path: Path) -> str:
    """Helper to get the content of readme in
    a CoSApp module

    Args:
        path (Path): Path to search for readme file

    Returns:
        str: Content of readme file
    """
    names = ["README", "readme"]
    exts = ["", "MD", "RST", "md", "rst", "TXT", "txt"]
    read_me = None
    for name in names:
        for ext in exts:
            if ext != "":
                file_path = path / f"{name}.{ext}"
            else:
                file_path = path / name
            if file_path.is_file():
                with open(file_path, "r") as f:
                    read_me = f.read()
                return read_me

    return read_me



import copy
from typing import Callable, Dict, List, Optional, Union

from cosapp.drivers.time.interfaces import ExplicitTimeDriver
from cosapp.systems import System
from abc import ABCMeta, abstractmethod


class CosappParser(metaclass=ABCMeta):
    """Abstract class to read/modifier/interact with  cosapp system

    Parameters
    ----------
    system : cosapp.systems.System
    """

    def __init__(self, system: Union[System, Dict]) -> None:
        self.system_name = None
        self._system = None

        self._children = {}
        self._driver = {}
        self.system_variable_data = {}
        self.variable_list = []

        self.set_up_system(system)

        self.flattened_system = self.flatten_system_dict(self.system_dict)
        self.tree_dict = self.system_to_tree_dict(self.system_dict)

    @abstractmethod
    def set_up_system(self, data: Union[Dict, System]) -> None:
        pass

    @property
    def children_list(self) -> List[str]:
        """Return the list of all sub-systems in the form
        of path (ex.: `parent.sub_parent.current_system`)
        """
        return list(self._children)

    @property
    def children_port(self) -> Dict[str, List[str]]:
        """Get a dictionary which contains system name
        as key and list of all ports of this system as value

        Returns
        -------
        Dict[str, List[str]]
            Dict contains name of port of each child in input system, including itself
        """
        ret = {}
        for sys, data in self._children.items():
            ret[sys] = data["port_list"]
        return ret

    @property
    def children_in_port(self) -> Dict[str, List[str]]:
        """Get a dictionary which contains system name
        as key and list of all input ports of this system as value

        Returns
        -------
        Dict[str, List[str]]
            Dict contains name of port of each child in input system, including itself
        """

        ret = {}
        for sys, data in self._children.items():
            ret[sys] = data["in_port_list"]
        return ret

    @property
    def children_out_port(self) -> Dict[str, List[str]]:
        """Get a dictionary which contains system name
        as key and list of all out ports of this system as value

        Returns
        -------
        Dict[str, List[str]]
            Dict contains name of port of each child in input system, including itself
        """
        ret = {}
        for sys, data in self._children.items():
            ret[sys] = data["out_port_list"]
        return ret

    @property
    def children_drive(self) -> Dict[str, List[str]]:
        """Get a dictionary which contains system name
        as key and list of all drivers of this system as value

        Returns
        -------
        Dict[str, List[str]]
            Dict contains name of drivers of each child in input system, including itself
        """
        ret = {}
        for key in self.children_list:
            if key in self._driver:
                driver = self._driver[key]
                ret[key] = list(driver) if len(driver) > 0 else ["None"]
        return ret

    def get_children_var(self, sys_name: str, port_name: str) -> List[str]:
        """Get all variables in a port of a system."""
        return []

    def get_children_var_input(
        self, sys_name: str, port_name: str, check_input: bool = False
    ) -> Dict[str, Dict]:
        """Get the variables of a port which is not a sink of any
        connection"""
        return {}

    def connect_main_driver(self, f: Callable) -> None:
        """Connect the computed signal of first driver of system to slot f"""
        pass

    def get_time_driver(self) -> List[ExplicitTimeDriver]:
        """Return the list of ExplicitTimeDriver in system"""
        return []

    @abstractmethod
    def _discover_children(
        self, system: Union[System, dict], parent: Optional[str] = None
    ) -> None:
        """Get the sub system of input system

        Parameters
        ----------
        system : Union[cosapp.systems.System, Dict]
            The input system

        parent: Optional[str]
            Path in form of "parent.child.sub-child ..." to the parent of input system.
            parent = None if input system does not have parent.

        Returns
        -------
        Dict[str, List[str]]
            Dictionary of sub system of input system with their children.
        """
        pass

    def flatten_system_dict(self, data: Dict) -> Dict:
        """Create of flattened version of self.system_dict with the keys
        in form of "parent.sub1.sub2" and content in form of
            {"class" : str,
            "inputs": {},
            "connections" : {},
            "subsystems" : List[str],
            "exec_order" : List[str]}

        Parameters
        ----------
        data : Dict
            System dictionary obtained by to_dict() method.

        Returns
        -------
        Dict[str, Dict]
            Flattend system dictionary.

        """

        flattened_system = {}
        if len(data) > 0:
            for system_path in self.children_list:
                system_list = system_path.split(".")
                if len(system_list) == 1:
                    current_system = data[system_list[0]]
                else:
                    current_system = data[system_list[0]]
                    for item in system_list[1:]:
                        current_system = current_system["subsystems"][item]

                flattened_system[system_path] = copy.deepcopy(current_system)
                try:
                    flattened_system[system_path]["subsystems"] = list(
                        current_system["subsystems"]
                    )
                except:
                    flattened_system[system_path]["subsystems"] = []

        return flattened_system

    def system_to_tree_dict(self, data: Dict, root=True, parent=None) -> Dict:
        """Convert the cosapp dict obtained by `to_dict`
        method into a compatible dict used by fontend to
        display the tree structure.

        Parameters
        ----------
        data : Dict
            System dictionary obtained by to_dict() method.

        root : boolean
            Flag to check if input data is main system or sub-system
        Returns
        -------
        Dict[str, Dict]
            System  tree dictionary.
        """

        ret = []
        if len(data) > 0:
            if root:
                temp = {
                    "title": self.system_name,
                    "id": self.system_name,
                    "expanded": True,
                }
                if "subsystems" in data[self.system_name]:
                    temp["children"] = self.system_to_tree_dict(
                        data[self.system_name]["subsystems"], False, self.system_name
                    )
                ret.append(temp)
            else:
                for key, key_data in data.items():
                    temp = {"title": key, "id": f"{parent}.{key}", "expanded": True}
                    if "subsystems" in key_data:
                        temp["children"] = self.system_to_tree_dict(
                            key_data["subsystems"], False, f"{parent}.{key}"
                        )
                    ret.append(temp)
        else:
            ret.append(
                {"title": self.system_name, "id": self.system_name, "expanded": True}
            )

        return ret



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.


import json
import os
from typing import Any, Dict, List, Type, Union
from weakref import ReferenceType, ref


from cosapp.systems import System

from traitlets import Unicode
from cosapp_lab.widgets.base import BaseWidget
from cosapp_lab.widgets.chartwidget import ChartElement
from cosapp_lab.widgets.controllerwidget import ControllerComponent
from cosapp_lab.widgets.geometrywidget import GeometryComponent
from cosapp_lab.widgets.structurewidget import StructureComponent
from cosapp_lab.widgets.infowidget import SystemInfoComponent
from .component import WidgetView, DocumentView, SysExplorerComponent


class SysExplorer(BaseWidget):

    _model_name = Unicode("SysExplorerModel").tag(sync=True)
    _view_name = Unicode("SysExplorerView").tag(sync=True)

    def __init__(self, data: Union[System, List[System]] = None, **kwargs):
        self.title = "SysExplorer widget"
        super().__init__(data, **kwargs)

    def init_component(self, **kwargs):
        self.register(SysExplorerComponent, **kwargs)
        self.register(ChartElement, **kwargs)
        self.register(StructureComponent, **kwargs)
        self.register(ControllerComponent, **kwargs)
        add_shape = kwargs.pop("add_shape", None)
        source = kwargs.pop("source", None)
        self.register(GeometryComponent, add_shape=add_shape, source=source, **kwargs)
        self.register(WidgetView, **kwargs)
        self.register(DocumentView, **kwargs)
        self.register(SystemInfoComponent, **kwargs)



import os
import pathlib

try:
    from cookiecutter.main import cookiecutter
except ImportError:
    cookiecutter = None


def cosapp_init():
    if cookiecutter is None:
        print("Please install cookiecutter before using cosapp init")
        return

    template_path = os.path.join(
        pathlib.Path(__file__).parent.absolute().parent,
        "utils",
        "templates",
        "cookiecutter-cosapp-workspace",
    )
    cookiecutter(template_path)



import os
from pathlib import Path
from tornado.web import RequestHandler
from enum import Enum
from jupyter_server.base.handlers import APIHandler


class MODULE_MODE(Enum):
    SINGLE = 0
    MULTIPLE = 1


ROOT = Path(os.path.dirname(__file__)).parent
TEMPLATE_ROOT = ROOT / "module" / "templates"
DEFAULT_STATIC_ROOT = ROOT / "app_static" / "static"


class StartupCodeHandler(APIHandler):
    """"""

    def initialize(self, **kwargs):
        self._code = kwargs.get("code", "")
        self._title = kwargs.get("title", "CoSApp application")
        self._args = kwargs.get("args", "[]")
        self._all_module_data = kwargs.get("all_module_data", {})

    def get(self):
        self.finish({"libData": self._all_module_data})

    def post(self):
        data = self.get_json_body()
        module_name = data.get("module", None)
        code = ""
        title = self._title
        if self._code is not None and len(self._code) > 0:
            code = self._code.replace("{{args}}", self._args)
        else:
            if module_name is not None and module_name in self._all_module_data:
                code = self._all_module_data[module_name]["code"].replace(
                    "{{args}}", self._args
                )
                title = self._all_module_data[module_name]["title"]
        self.finish({"code": code, "title": title})


class MainModuleHandler(RequestHandler):
    """"""

    def initialize(self, **kwargs):
        self._all_module_data = kwargs.get("all_module_data", {})
        self._module_mode = kwargs.get("module_mode", MODULE_MODE.MULTIPLE)

    def get(self, module_name):
        static_prefix = ".."
        if self._module_mode == MODULE_MODE.SINGLE:
            self.render(
                str(TEMPLATE_ROOT / "404.html"),
                error_msg="CoSApp Lab server started in single module mode!",
                static_prefix=static_prefix,
            )
        else:
            if module_name not in self._all_module_data:
                self.render(
                    str(TEMPLATE_ROOT / "404.html"),
                    error_msg="Module can not be found!",
                    static_prefix=static_prefix,
                )
            else:
                path = ROOT / "app_static" / "main.html"
                self.render(str(path), static_prefix=static_prefix)


class MainEntryHandler(RequestHandler):
    """"""

    def initialize(self, **kwargs):
        self._all_module_data = kwargs.get("all_module_data", {})
        self._module_mode = kwargs.get("module_mode", MODULE_MODE.MULTIPLE)

    def get(self):
        if self._module_mode == MODULE_MODE.SINGLE:
            path = ROOT / "app_static" / "main.html"
            static_prefix = ""
        elif self._module_mode == MODULE_MODE.MULTIPLE:
            path = ROOT / "app_static" / "index.html"
            static_prefix = ""
        self.render(str(path), static_prefix=static_prefix)


class Default404Handler(RequestHandler):
    def prepare(self):
        self.set_status(404)
        self.render(
            str(TEMPLATE_ROOT / "404.html"),
            error_msg="Invalid resource path!",
            static_prefix="..",
        )



from typing import Any, Dict

from ipywidgets import Layout, VBox

from cosapp.systems import System

from .controller import Controller
from .validation import Validation


class SideBar(VBox):
    def __init__(self, system: System, definition: Dict[str, Dict[str, Any]]):
        controls = Controller(system, definition)
        validation = Validation(system)
        super(SideBar, self).__init__(
            children=[controls, validation],
            layout=Layout(align_items="center", width="100%", min_width="282px"),
        )



import click
from click.utils import echo
from cosapp_lab.script import cosapp_script, LOGO
from cosapp_lab import __version__ as cosapp_lab_version
from cosapp.core import __version__ as cosapp_version

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(
    invoke_without_command=True,
    no_args_is_help=False,
    context_settings=CONTEXT_SETTINGS,
)
@click.option("-v", "--version", is_flag=True, default=False)
@click.pass_context
def cosapp(ctx, version):
    """CosApp CLI tool."""
    echo(LOGO)
    if version:
        echo(f"CosApp version {cosapp_version}")
        echo(f"CosApp Lab version {cosapp_lab_version}")
    else:
        if ctx.invoked_subcommand is None:
            lab_app = cosapp_script["lab"]
            if lab_app is not None:
                lab_app()
            else:
                echo(
                    f"Please start CoSApp in an environnement with JupyterLab installed "
                )


@cosapp.command("init", short_help="Initialize a new CosApp project")
def init():
    cosapp_script["init"]()


@cosapp.group(
    "module",
    short_help="CoSApp modules management tool",
    invoke_without_command=False,
    no_args_is_help=False,
    context_settings=CONTEXT_SETTINGS,
)
def cosapp_module():
    pass


@cosapp_module.command(
    "register", short_help="Register a CoSApp library as standalone module"
)
@click.argument("name")
def cosapp_module_register(name):
    cosapp_script["cosapp_module_register"]([name])


@cosapp_module.command(
    "remove", short_help="Remove a CoSApp library from standalone module list"
)
@click.argument("module")
def cosapp_module_register(module):
    cosapp_script["cosapp_module_remove"]([module])


@cosapp_module.command("list", short_help="List all CoSApp standalone modules")
def cosapp_module_register():
    cosapp_script["cosapp_module_list"]()


@cosapp_module.command("update", short_help="Update all modules configuration")
def cosapp_module_register():
    cosapp_script["cosapp_module_update"]()


@cosapp.command("load", short_help="Start a CosApp module in standalone mode")
@click.argument("module", default=None, required=False)
@click.option(
    "-a",
    "--all",
    flag_value=True,
    required=False,
    help="Flag to load all available module",
)
@click.option("-f", "--file", default=None, required=False, help="Path to script file")
@click.option(
    "-args",
    "--arguments",
    default=None,
    required=False,
    nargs=1,
    type=str,
    help="A string contains list of parameters for module initial function",
)
@click.option(
    "-p",
    "--port",
    default=6789,
    required=False,
    nargs=1,
    type=int,
    help="Port number for CoSApp Lab server",
)
def load_module(module, file, all, arguments, port):
    """Start a CoSApp standalone module. Use -f/--file option to start server from a python script

    - MODULE: named of registered CoSApp module.
    """
    if all:
        cosapp_script["cosapp_load_module"]("__all__", arguments)
    else:
        if module is None and file is None:
            echo("Please specify module name or file path.")
            return
        elif module is None:
            if arguments is not None:
                echo("Arguments is not supported for python script")
            cosapp_script["cosapp_load_file"](file)
        elif file is None:
            cosapp_script["cosapp_load_module"](module, arguments)
        else:
            cosapp_script["cosapp_load_module"](module, arguments)


if __name__ == "__main__":
    cosapp()  # pragma: no cover



import re
from weakref import ref

from ipywidgets import HTML, HBox, Layout, ToggleButton, VBox
from traitlets import Dict, Unicode

from cosapp.systems import System
from cosapp.utils.validate import validate


class Validation(VBox):
    warnings = Dict(value_trait=Unicode())
    errors = Dict(value_trait=Unicode())

    CLEAN_NAME = r"\.(" + System.INWARDS + r"|" + System.OUTWARDS + r")\."

    def __init__(self, system: System):
        self._system = ref(system)
        self.status_button = ToggleButton(
            description="Nothing run yet",
            button_style="",
            tooltip="Click me to show the log.",
            layout=Layout(height="100%"),
        )
        self.validation_log = HTML(
            value="", layout=Layout(display="none", overflow_y="auto")
        )
        super(Validation, self).__init__(
            children=[
                HBox([self.status_button], layout=Layout(justify_content="center")),
                self.validation_log,
            ],
            layout=Layout(align_self="flex-start", width="100%"),
        )

        # Connect event
        def toggle(change):
            self.validation_log.layout.display = "unset" if change["new"] else "none"

        self.status_button.observe(toggle, names="value")

        # Add callback to system
        system.clean_ran.connect(self._set_validities)

        # Update status
        self._set_validities()

    def _set_validities(self):
        """Update UI depending on the validation status."""
        if self._system() is None:
            return

        self.warnings, self.errors = validate(self._system())

        warnings = {
            re.sub(Validation.CLEAN_NAME, ".", name): comment
            for name, comment in self.warnings.items()
        }
        errors = {
            re.sub(Validation.CLEAN_NAME, ".", name): comment
            for name, comment in self.errors.items()
        }

        # Feedback on button
        style = "success"
        info = "Ok"

        # Customize here the HTML rendering
        log_html = """<div id="validationLog">
            <style>
                #validationLog { 
                    white-space: nowrap;
                    line-height: normal;
                }
            </style>"""
        if len(errors):
            style = "danger"
            info = "Out of physical limits"
            log_html += "<h3>Out of modeling limits</h3><ul>"

            for name, comment in errors.items():
                log_html += "<li>{}{}</li>".format(name, comment)
            log_html += "</ul>"

        if len(warnings):
            if style != "danger":
                style = "warning"
                info = "On modeling limits"
            log_html += "<h3>On modeling limits</h3><ul>"

            for name, comment in warnings.items():
                log_html += "<li>{}{}</li>".format(name, comment)
            log_html += "</ul>"
        log_html += "</div>"

        self.status_button.button_style = style
        self.status_button.description = info
        self.validation_log.value = log_html



from .sysexplorer import SysExplorer
from .chartwidget.chart_widget import ChartWidget
from .datawidget.data_widget import DataWidget
from .controllerwidget.controller_widget import ControllerWidget
from .geometrywidget.geometry_widget import GeometryWidget
from .pbswidget.pbs_widget import PbsWidget
from .connectionwidget.connection_widget import ConnectionWidget
from .structurewidget.structure_widget import StructureWidget
from .infowidget.info_widget import SystemInfoWidget
from .legacy.base import NumberData, StringData, EnumData, ArrayData
from .legacy.widgetlogger import WidgetLogger
from .legacy.sidebar import SideBar
from .legacy.data import Scatter
from .legacy.statistics import Statistics


__all__ = [
    "SysExplorer",
    "ChartWidget",
    "DataWidget",
    "ControllerWidget",
    "GeometryWidget",
    "PbsWidget",
    "ConnectionWidget",
]



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.


from typing import Any, Dict, List, Type, Union

from cosapp.systems import System
from cosapp_lab._frontend import module_name, module_version
from cosapp_lab.widgets.utils import (
    CosappObjectParser,
    is_jsonable,
)
import copy, json
from ipywidgets import Box
from traitlets import CaselessStrEnum
from traitlets import Unicode
from traitlets import Dict as tDict
from traitlets import Int, Unicode
from weakref import ReferenceType, ref
from .base_component import BaseComponent
import numpy as np


class BaseWidget(Box):

    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)
    title = Unicode("BaseWidget").tag(sync=True)
    anchor = CaselessStrEnum(
        [
            "split-right",
            "split-left",
            "split-top",
            "split-bottom",
            "tab-before",
            "tab-after",
            "right",
        ],
        default_value="split-right",
        allow_none=False,
    ).tag(sync=True)
    system_data = tDict(default_value={"key": "None"}, allow_none=False).tag(sync=True)
    update_signal = Int(default_value=0, allow_none=False).tag(sync=True)

    initial_store = tDict(default_value={}, allow_none=False).tag(sync=True)
    chart_template = tDict(default_value={}, allow_none=False).tag(sync=True)

    system_config = tDict(default_value={}, allow_none=False).tag(sync=True)

    def __init__(self, data: Union[System, List[System]] = None, **kwargs):

        template_path = kwargs.pop("template", None)

        if template_path is not None:
            try:
                with open(template_path, "r") as f:
                    json_data = json.load(f)
                    if "modelJson" not in json_data:
                        raise KeyError(
                            f"Required keys can not be found in {template_path}"
                        )
                    else:
                        self.chart_template = {
                            "chart_template": json_data,
                            "template_path": template_path,
                        }
                        self._template_path = template_path.replace(".json", "")
            except FileNotFoundError:
                raise FileNotFoundError(f"Can not read {template_path}")
        else:
            self.chart_template = {}
            self._template_path = ""
        if hasattr(data, "__iter__"):
            root = System("chart_viewer")
            for sys in data:
                root.add_child(sys)
                key_list = list(sys.drivers)
                if len(key_list) > 0:
                    sys.drivers[key_list[0]].computed.connect(
                        self.computed_notification
                    )
                else:
                    sys.computed.connect(self.computed_notification)

            self.init_value(root, **kwargs)
        else:
            self.init_value(data, **kwargs)

        self.__component = {}
        self.msg_handlers = []
        self.computed_callbacks = []
        self.init_component(_template_path=self._template_path, **kwargs)
        super().__init__(**kwargs)

    def init_component(self, **kwargs):
        raise NotImplementedError

    def init_value(self, data: System, **kwargs):
        self._system: ReferenceType[System] = ref(data)
        self.system_config = {"mode": "run", "enableEdit": True, "root_name": data.name}
        user_config = kwargs.get("config", {})
        for key, val in user_config.items():
            if key in self.system_config:
                self.system_config[key] = val
        self.initial_store = {"systemConfig": copy.deepcopy(self.system_config)}
        self.update_signal = 0
        self.time_step = 0

        self.__init_data(data)
        self.__init_connection()

    def __init_data(self, data: System):
        """Parse cosapp System and return list of sub-systems,
        list of ports...

        Parameters
        ----------
        data : cosapp.systems.System
            Input system
        """
        self.system_variable = {}
        if isinstance(data, System):
            self.sys_data = CosappObjectParser(data)
        else:
            raise TypeError(f"{data} is not a CoSApp system")
        self.system_dict = self.sys_data.flattened_system
        # self._system = ref(data)
        self._system_list = self.sys_data.children_list
        self._driver_list = self.sys_data.children_drive
        in_port_dict = self.sys_data.children_in_port
        out_port_dict = self.sys_data.children_out_port
        self.systemGraphData = {}
        for key in self._system_list:
            try:
                connection_list = self.system_dict[key]["connections"]
            except:
                connection_list = []
            self.systemGraphData[key] = {
                "inPort": in_port_dict[key],
                "outPort": out_port_dict[key],
                "connections": connection_list,
            }

        port_dict = self.sys_data.children_port
        for sys_name in port_dict:
            for port_name in port_dict[sys_name]:
                variable_dict = self.sys_data.get_children_var_input(
                    sys_name, port_name
                )
                if "Mutable variable not found" not in variable_dict:
                    for var_name in variable_dict:
                        if isinstance(variable_dict[var_name]["value"], np.ndarray):
                            variable_dict[var_name]["value"] = variable_dict[var_name][
                                "value"
                            ].tolist()
                            self.system_variable[
                                f"{sys_name}.{port_name}.{var_name}"
                            ] = variable_dict[var_name]
                        elif is_jsonable(variable_dict[var_name]["value"]):
                            self.system_variable[
                                f"{sys_name}.{port_name}.{var_name}"
                            ] = variable_dict[var_name]

        computedResult = self.sys_data.serialize_data_from_system(False)
        recorderData = self.sys_data.serialize_recorder()
        driverData = self.sys_data.serialize_driver_data()

        self.system_data = {
            "systemGraph": {
                "systemGraphData": self.systemGraphData,
                "systemList": self._system_list,
                "graphJsonData": {},
            },
            "systemPBS": {},
            "systemTree": self.sys_data.tree_dict,
            "portMetaData": self.sys_data.children_port_meta,
            "variableData": self.system_variable,
            "computedResult": computedResult,
            "recorderData": recorderData,
            "driverData": driverData,
        }

    def __init_connection(self) -> None:
        """Initialize the connection between fontend - backend and
        between input system with the callbacks.
        """
        self.sys_data.connect_main_driver(self.computed_notification)
        self.on_msg(self._handle_button_msg)

    def computed_notification(self) -> None:
        """Callback function used to update geometry data from buffer
        and to emit update signal to front end.

        """
        for callback in self.computed_callbacks:
            callback()

    def _handle_button_msg(self, model: Any, content: Dict, buffers: List) -> None:
        """Helper function to receive message from front end.

        Parameters
        ----------
        model : ChartViewerModel
            Current model value.

        content : {"action" : "runSignal" | "requestUpdate", "payload" : Dict}
            If action == "runSignal" => run the main
            driver of system, payload contains a dict of system parameters
            and its value.
            If action == "requestUpdate" => Send the geometry of
            the last computed step to front-end to update the computing
            progress.

        buffers : List

        """
        for msg_handler in self.msg_handlers:
            msg_handler(model, content, buffers)

    def register(self, Cls: Type[BaseComponent], **kwargs) -> None:
        component = Cls(self._system, self.sys_data, self.send, **kwargs)
        if Cls.name not in self.__component:
            self.__component[Cls.name] = component
        else:
            raise KeyError("Component already registed")

        self.msg_handlers.append(
            component._handle_button_msg
        )  # TODO Use dict instead of list
        self.computed_callbacks.append(component.computed_notification)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.


from typing import Any, Callable, Dict, List
from weakref import ReferenceType
import os, json
from cosapp.systems import System
from cosapp_lab.widgets.base.base_component import BaseComponent
from cosapp_lab.widgets.utils import (
    CosappObjectParser,
    get_nonexistant_path,
)


class SysExplorerComponent(BaseComponent):
    name = "SysExplorer"

    def __init__(
        self,
        data: "ReferenceType[System]",
        sys_data: CosappObjectParser,
        send_func: Callable,
        **kwargs,
    ):
        self._template_path = kwargs.get("_template_path")
        super().__init__(data=data, sys_data=sys_data, send_func=send_func, **kwargs)

    def _handle_button_msg(self, model: Any, content: Dict, buffers: List) -> None:

        if content["action"] == "SysExplorer::chartViewerSaveJson":
            json_name = content["payload"]["jsonName"].replace(".json", "")
            if json_name != self._template_path:
                file_path = get_nonexistant_path(
                    os.path.join(os.getcwd(), f"{json_name}.json")
                )
                self._template_path = os.path.basename(file_path).replace(".json", "")
            else:
                file_path = f"{json_name}.json"
            self.send(
                {
                    "type": "SysExplorer::update_save_path",
                    "payload": {"templatePath": self._template_path},
                }
            )
            json_data = content["payload"]["jsonData"]
            with open(file_path, "w") as f:
                json.dump(json_data, f)



from cosapp_lab.widgets.base.base_widget import BaseWidget
from typing import List, Union
from cosapp.systems import System
from traitlets import Unicode
from .controller_component import ControllerComponent


class ControllerWidget(BaseWidget):

    _model_name = Unicode("ControllerWidgetModel").tag(sync=True)
    _view_name = Unicode("ControllerWidgetView").tag(sync=True)

    def __init__(self, data: Union[System, List[System]] = None, **kwargs):
        self.title = "Controller widget"
        super().__init__(data, **kwargs)

    def init_component(self, **kwargs):
        self.register(ControllerComponent, **kwargs)



from enum import Enum
from typing import Any, Dict, Iterable, List, Optional, Union

import numpy as np
import pandas as pd

try:
    import plotly.graph_objs as go
except:
    go = None
from ipywidgets import Button, Dropdown, HBox, HTML, IntSlider, Layout, Output, VBox
from traitlets import List, Unicode, observe

from cosapp.drivers import RunOnce
from cosapp.drivers.abstractsolver import AbstractSolver
from cosapp.recorders import DataFrameRecorder
from cosapp.utils.helpers import is_number

from .sidecar import Sidecar


class DataFrameRecorderViewer(VBox):
    cases = List(trait=Unicode())
    variables = List(trait=Unicode())

    def __init__(self, rec: DataFrameRecorder, **kwargs) -> None:
        super(DataFrameRecorderViewer, self).__init__(**kwargs)
        self.recorder = rec
        self._extra_data = []

        vars_options = [
            var
            for var in self.recorder.field_names()
            if var in self.recorder._watch_object
            and is_number(self.recorder._watch_object[var])
        ]
        self.variables.append("index")
        self.variables.extend(vars_options)

        self.cases.append("all")
        if isinstance(self.recorder._owner, AbstractSolver):
            self.cases.extend(
                [
                    name
                    for name, d in filter(
                        lambda n_d: isinstance(n_d[1], RunOnce),
                        self.recorder._owner.children.items(),
                    )
                ]
            )

        # Connect with recorder
        # Must be a method as local function will be clean by weakref in Slot
        rec.state_recorded.connect(self._soft_update)
        rec.cleared.connect(self.update)

    def _soft_update(self, change: Dict[str, Any] = {"new": True}, **kwargs):
        """Call `update` with `force=False`."""
        self.update(force=False)

    def update(
        self, change: Dict[str, Any] = {"new": True}, force: bool = True, **kwargs
    ) -> None:
        """Handler to update the figure."""
        if self.recorder.paused:
            return

        # Append new extra data if there are some
        if self.recorder.extra_data != self._extra_data:
            self._extra_data = self.recorder.extra_data
            self.variables.extend(
                [f for f in self._extra_data if f not in self.variables]
            )


class Scatter(DataFrameRecorderViewer):
    """Notebook widget to handle a scatter plot with data obtained from a DataFrameRecorder

    Parameters
    ----------
    rec : cosapp.recorder.DataFrameRecorder
        Recorder containing the data to view
    """

    def __init__(self, rec: DataFrameRecorder, **kwargs) -> None:
        if go is None:
            raise ImportError(" this widget requires plotly")

        super(Scatter, self).__init__(rec, **kwargs)

        self.widgets = w = {
            "restorer": {
                "button": Button(
                    icon="undo",
                    layout=Layout(width="auto", flex="0 0 auto", align_items="center"),
                    width="auto",
                    disabled=True,
                ),
                "slider": IntSlider(
                    min=0, max=0, layout=Layout(flex="1 1 auto"), continuous_update=True
                ),
            },
            "plot": self._init_plot(self.recorder._owner.name),
            "axes": {
                "case": Dropdown(
                    options=self.cases,
                    description="case",
                    disabled=len(self.cases) <= 1,
                ),
                "x": Dropdown(options=self.variables, description="x", disabled=False),
                "y": Dropdown(options=self.variables, description="y", disabled=False),
            },
            "debug": Output(),
        }

        # Connect internal widgets
        w["restorer"]["slider"].observe(self.index_handler, "value")
        w["restorer"]["button"].on_click(self.button_action)
        w["axes"]["case"].observe(self.case_handler, "value")
        w["axes"]["x"].observe(self.x_handler, "value")
        w["axes"]["y"].observe(self.y_handler, "value")

        self.layout = Layout(
            height="100%", margin="5px", padding="0 0 10px 0", overflow_y="hidden"
        )
        self.children = [
            HBox(
                [
                    w["restorer"]["button"],
                    w["restorer"]["slider"],
                    HTML(
                        "<style>.cosapp-sidecar .js-plotly-plot { height: 100%; }</style>"
                    ),
                ],
                layout=Layout(
                    flex="0 0 auto",
                    justify_content="flex-end",
                    align_items="center",
                    min_height="30px",
                    overflow_y="hidden",
                ),
            ),
            w["plot"],
            HBox(
                [w["axes"]["case"], w["axes"]["x"], w["axes"]["y"]],
                layout=Layout(
                    justify_content="center",
                    width="auto",
                    min_height="30px",
                    overflow_y="hidden",
                    flex="1 1 auto",
                ),
            ),
            VBox([w["debug"]], layout=Layout(overflow_y="auto")),
        ]

        # Force plot update
        self.update()

    def _init_plot(self, name: str) -> "plotly.graph_objs.FigureWidget":
        """Build and display the Plotly plot in the Jupyterlab front-end.

        Parameters
        ----------
        name : str
            Figure title
        """

        scatter = go.Scatter(
            x=[], y=[], mode="markers", marker=dict(size=8, color="grey")
        )
        layout = go.Layout(
            title=name,
            hovermode="closest",
            margin=go.layout.Margin(l=70, r=50, b=40, t=50, pad=1),
            xaxis=dict(
                showgrid=True,
                zeroline=True,
                showline=True,
                nticks=10,
                hoverformat=".3f",
                title="",
            ),
            yaxis=dict(
                showgrid=True,
                zeroline=True,
                showline=True,
                nticks=5,
                hoverformat=".3f",
                title="",
            ),
        )
        return go.FigureWidget(data=[scatter], layout=layout)

    @observe("variables")
    def on_variables(self, change):
        """Handle changes in variable list."""
        for axis in ("x", "y"):
            dropdown = self.widgets["axes"][axis]
            v = dropdown.value
            dropdown.options = change["new"]
            dropdown.value = v

    def case_handler(self, change: Dict[str, Any]) -> None:
        """Handle *case* combobox change."""
        new = change["new"]
        if new:
            self.update({"case": new})

    def x_handler(self, change: Dict[str, Any]) -> None:
        """Handle *x* combobox change."""
        new = change["new"]
        if new:
            self.update({"x": new})

    def y_handler(self, change: Dict[str, Any]) -> None:
        """Handle *y* combobox change."""
        new = change["new"]
        if new:
            self.update({"y": new})

    def index_handler(self, change: Dict[str, Any]) -> None:
        """Handle *index* slider change."""
        new = change["new"]
        if new:
            self.update({"index": new})

    def button_action(self, action) -> None:
        """Handle click event on restore button."""
        self.recorder.paused = True
        self.recorder._watch_object.run_drivers()
        self.recorder.paused = False

    def update(
        self, change: Dict[str, Any] = {"new": True}, force: bool = True, **kwargs
    ) -> None:
        """Handler to update the figure."""
        if self.recorder.paused:
            return

        with self.widgets["debug"]:

            super(Scatter, self).update(change, force, **kwargs)

            if change:
                case_axis = self.widgets["axes"]["case"].value
                n_cases = max(1, len(self.cases) - 1) if case_axis == "all" else 1
                x_axis = self.widgets["axes"]["x"].value
                y_axis = self.widgets["axes"]["y"].value
                data = self.recorder.data

                if case_axis != "all" and data.shape[0] > 0:
                    if case_axis in data["Reference"].unique():
                        data = data.loc[data["Reference"] == case_axis]
                    else:
                        data = pd.DataFrame(columns=data.columns)
                count = int(data.shape[0] / n_cases)

                if "index" not in change:
                    current_idx = max(count - 1, 0)
                    self.widgets["restorer"]["slider"].max = current_idx
                    self.widgets["restorer"]["slider"].unobserve(
                        self.index_handler, names="value"
                    )
                    self.widgets["restorer"]["slider"].value = current_idx
                    self.widgets["restorer"]["slider"].observe(
                        self.index_handler, names="value"
                    )
                    self.widgets["restorer"]["button"].disabled = count <= 1
                else:
                    current_idx = self.widgets["restorer"]["slider"].value
                    self.recorder.restore(current_idx)

                colors = ["grey"] * count * n_cases
                opacities = [0.3] * count * n_cases
                if count > 0:
                    for data_idx in range(
                        current_idx * n_cases, (current_idx + 1) * n_cases
                    ):
                        colors[data_idx] = "orange"
                        opacities[data_idx] = 1.0

                src = {
                    "x": data[x_axis].values if x_axis != "index" else np.arange(count),
                    "y": data[y_axis].values if y_axis != "index" else np.arange(count),
                    "text": list(range(count)),
                    "marker": {"color": colors, "opacity": opacities},
                }

                if "new" not in change:
                    if "case" not in change:
                        if "x" not in change:
                            src.pop("x")
                        if "y" not in change:
                            src.pop("y")
                    if "index" not in change:
                        src.pop("marker")

                self.widgets["plot"].update(
                    data=[src],
                    layout={"xaxis": {"title": x_axis}, "yaxis": {"title": y_axis}},
                )



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.


"""
Widget container as geometry viewer panel
"""
from collections import OrderedDict
from typing import Any, Callable, Dict, List, Union
from weakref import ReferenceType

from cosapp.recorders import DataFrameRecorder
from cosapp.systems import System
from cosapp_lab.widgets.utils import CosappObjectParser, OccParser
from cosapp_lab.widgets.base.base_component import BaseComponent


class GeometryComponent(BaseComponent):
    name = "GeometryView"

    def __init__(
        self,
        data: "ReferenceType[System]",
        sys_data: CosappObjectParser,
        send_func: Callable,
        add_shape=None,
        source=None,
        **kwargs
    ):
        super().__init__(data=data, sys_data=sys_data, send_func=send_func, **kwargs)

        self.add_shape = add_shape
        self.geo_source = source
        self.time_step = 0
        self.__init_connection()

    def __init_connection(self):
        """Initialize the connection between fontend - backend and
        between input system with the callbacks.
        """
        if self.geo_source is not None and "recorder" in self.geo_source:
            self._static = False
            path_list = self.geo_source["recorder"].split(".")
            driver_name = path_list[-1]
            if len(path_list) == 1:
                recorder = self.system.drivers[driver_name].recorder
            else:
                recorder = (
                    self.system[".".join(path_list[0:-1])]
                    .drivers[driver_name]
                    .recorder
                )
            recorder.state_recorded.connect(self.save_data)
        else:
            driver_list = self.sys_data.get_time_driver()
            if len(driver_list) > 0:
                self._static = False
                time_driver = driver_list[0]
                if time_driver.recorder is not None:
                    time_driver.recorder.state_recorded.connect(self.save_data)
                else:
                    notification_recorder = time_driver.add_recorder(
                        DataFrameRecorder(includes=["_"])
                    )
                    notification_recorder.state_recorded.connect(self.save_data)
            else:
                self._static = True

    def save_data(self, **kwarg) -> None:
        """Callback function used to get geometry data
        after each time step.
        """
        data: OccParser = self.get_geometry(self.system)
        threejs_data = data.threejs_data
        binary_data = []
        binary_position = data.binary_position

        for b_data in data.binary_data:
            binary_data.append(b_data.tobytes())
        payload = {
            "threejs_data": threejs_data,
            "binary_position": binary_position,
            "time_step": self.time_step,
        }
        self.send({"type": "GeometryView::geo_data", "payload": payload}, binary_data)
        self.time_step += 1

    def send_initial_geometry(self) -> None:
        if self.geo_source is not None and "recorder" in self.geo_source:
            all_data = self.get_geometry(self.system, get_all=True)
            payload_length = len(all_data)
            for idx in all_data:
                payload_length -= 1
                threejs_data = all_data[idx].threejs_data
                binary_data = []
                binary_position = all_data[idx].binary_position
                for b_data in all_data[idx].binary_data:
                    binary_data.append(b_data.tobytes())
                payload = {
                    "threejs_data": threejs_data,
                    "binary_position": binary_position,
                    "time_step": idx,
                    "remaining": payload_length,
                }
                self.send(
                    {"type": "GeometryView::geo_data", "payload": payload}, binary_data
                )
        else:

            data = self.get_geometry(self.system)
            threejs_data = data.threejs_data
            binary_data = []
            binary_position = data.binary_position
            if len(threejs_data) > 0:
                for b_data in data.binary_data:
                    binary_data.append(b_data.tobytes())
                payload = {
                    "threejs_data": threejs_data,
                    "binary_position": binary_position,
                    "time_step": 0,
                    "remaining": 0,
                }
                self.send(
                    {"type": "GeometryView::geo_data", "payload": payload}, binary_data
                )

    def get_geometry(
        self, sys: System, get_all=False
    ) -> Union[
        OccParser, Dict[int, OccParser]
    ]:  # Dict[int, List[Dict[str, List[Union[int, float]]]]]:
        """Convert the open cascade object inside system into
        serializable data in order to send to front end.

        Parameters
        ----------
        sys : System
            The input system.

        Returns
        -------
        Dict[int, List[Dict[str, List[Union[int, float]]]]]
            A dictionary contains geometry data.
        """

        if self.geo_source is not None and "recorder" in self.geo_source:

            path_list = self.geo_source["recorder"].split(".")
            driver_name = path_list[-1]
            if len(path_list) == 1:
                recorder = self.system.drivers[driver_name].recorder
            else:
                recorder = (
                    self.system[".".join(path_list[0:-1])]
                    .drivers[driver_name]
                    .recorder
                )
            try:
                recorder_data = recorder.export_data()
            except:
                recorder_data = recorder.data
            n_index = len(recorder_data.index)
            if get_all:
                ret = OrderedDict()
                for idx in range(n_index):
                    r = []
                    for var_name in self.geo_source.get("variables", []):
                        column = recorder_data.get(var_name)
                        r.append(column[idx])
                    shape_data = OccParser(r)
                    ret[idx] = shape_data
                return ret
            else:
                r = []
                for var_name in self.geo_source.get("variables", []):
                    column = recorder_data.get(var_name)
                    r.append(column[n_index - 1])
                shape_data = OccParser(r)
                return shape_data

        else:

            r = []
            if self.add_shape is not None:

                self.add_shape(r, sys)
            else:
                try:
                    GeometryComponent.default_add_shape(r, sys)
                except:
                    pass

            shape_data = OccParser(r)
            return shape_data

    @staticmethod
    def default_add_shape(r: List, system: System) -> None:
        """Helper function used to extract all geometry inside
        system, it will be used if user defined function is not provided.

        Parameters
        ----------
        r : List
            The list contains all geometry variable of system.

        system : System
            The input system.

        """
        # TODO The method used to get shape from system should be generalized
        # in CosApp level instead of interface level.

        for child in system.children.values():
            GeometryComponent.default_add_shape(r, child)

        def filter_shapes(p: "ExtensiblePort") -> "bool":
            return (
                p.__class__.__name__ == "GeometryPort"
                and p.visible
                and p.shape is not None
            )

        for port in filter(filter_shapes, system.outputs.values()):
            if isinstance(port.shape, List):
                r.extend(port.shape)
            else:
                r.append(port.shape)

    def computed_notification(self) -> None:
        """Callback function used to update geometry data from buffer
        and to emit update signal to front end.
        """

        if self._static:
            data = self.get_geometry(self.system)
            threejs_data = data.threejs_data
            binary_data = []
            binary_position = data.binary_position
            if len(threejs_data) > 0:
                for b_data in data.binary_data:
                    binary_data.append(b_data.tobytes())
                payload = {
                    "threejs_data": threejs_data,
                    "binary_position": binary_position,
                    "time_step": 0,
                    "remaining": 0,
                }
                self.send(
                    {"type": "GeometryView::geo_data", "payload": payload}, binary_data
                )
        else:
            self.send(
                {
                    "type": "GeometryView::update_signal",
                    "payload": {},
                }
            )
            self.time_step = 0

    def _handle_button_msg(self, model: Any, content: Dict, buffers: List):
        if content["action"] == "GeometryView::requestInitialGeometry":
            self.send_initial_geometry()



import logging
from typing import Any, Dict, Iterable, Tuple, Union
from weakref import ref

from cosapp.systems import System
from ipywidgets import Accordion, Button, HBox, Layout, ToggleButton, VBox, Widget

from .base import SystemWidget

logger = logging.getLogger(__name__)


class Controller(VBox):
    """Widget allowing modification of parameters.

    This widget is composed of a header with buttons and accordions containing data handlers.

    The definition of the handlers is a dictionary of handlers list; the dictionary keys being
    the accordion group title.
    And handler can be defined by either:

    - A string: the name of the variable to handle
    - Two strings: the name of the variable to hanlde and the alias name to show
    - A dictionary: the mandatory argument is "name" but other can be added depending on the
        variable type
    - A `SystemWidget`: a widget to insert

    Parameters
    ----------
    system : cosapp.systems.System
        System to be controlled
    definition : Dict[str, Iterable[Union[str, Dict[str, Any], Tuple[str, str], SystemWidget]]]
        Definition of controllers
    """

    def __init__(
        self,
        system: System,
        definition: Dict[
            str, Iterable[Union[str, Dict[str, Any], Tuple[str, str], SystemWidget]]
        ],
    ):
        self._system = ref(system)
        self._controls = None
        super(Controller, self).__init__(
            children=self._build_widgets(system, definition),
            layout=Layout(align_self="flex-start", width="100%"),
        )

    def _build_widgets(
        self,
        system: System,
        definition: Dict[
            str, Iterable[Union[str, Dict[str, Any], Tuple[str, str], SystemWidget]]
        ],
    ) -> Iterable[Widget]:
        """Build the controller content.

        Parameters
        ----------
        system : cosapp.systems.System
            System to be controlled
        definition : Dict[str, Iterable[Union[str, Dict[str, Any], Tuple[str, str], SystemWidget]]]
            Definition of controllers
        """
        play_but = Button(
            icon="play",
            tooltip="Execute {}".format(system.name),
            layout=Layout(width="auto"),
        )
        pause_but = ToggleButton(
            value=not SystemWidget.auto_run,
            icon="pause",
            tooltip="Toogle system auto-execution.",
            layout=Layout(width="auto"),
        )
        undo_but = Button(
            icon="undo", tooltip="Reset to initial value", layout=Layout(width="auto")
        )
        header = HBox([play_but, pause_but, undo_but])

        groups = dict()
        for grp_name, group in definition.items():
            groups[grp_name] = list()
            for element in group:
                if isinstance(element, SystemWidget):
                    groups[grp_name].append(element)
                else:
                    alias = None
                    kwargs = dict()
                    if isinstance(element, str):
                        name = element
                    elif isinstance(element, dict):
                        name = element.pop("name")
                        alias = element.pop("alias", None)
                        kwargs = element
                    elif len(element) == 2:
                        name, alias = element
                    else:
                        raise ValueError(
                            "Definition '{}' is not valid.".format(element)
                        )
                    if alias is None and name.startswith(grp_name + "."):
                        alias = name[len(grp_name) + 1 :]
                    groups[grp_name].append(
                        SystemWidget.get_from_name(system, name, alias=alias, **kwargs)
                    )
        self._controls = Accordion(
            children=[VBox(groups[grp_name]) for grp_name in groups]
        )
        for i, grp_name in enumerate(groups):
            self._controls.set_title(i, grp_name)

        children = [header, self._controls]

        # Connect to events
        play_but.on_click(self.execute)

        def toogle_pause(change):
            SystemWidget.auto_run = not change["new"]

        pause_but.observe(toogle_pause, names="value")
        undo_but.on_click(self.reset)

        return children

    def execute(self, event=None):
        """Callback function for play button."""
        s = self._system()
        if s is not None:
            try:
                s.run_drivers()
            except Exception as e:
                logger.exception(e)

    def reset(self, event=None):
        """Callback function for reset button."""
        for group in self._controls.children:
            for widget in group.children:
                widget.reset(event)



from cosapp_lab.widgets.base.base_widget import BaseWidget
from typing import List, Union
from cosapp.systems import System
from traitlets import Unicode
from .chart_component import ChartElement


class ChartWidget(BaseWidget):

    _model_name = Unicode("ChartWidgetModel").tag(sync=True)
    _view_name = Unicode("ChartWidgetView").tag(sync=True)

    def __init__(self, data: Union[System, List[System]] = None, **kwargs):
        self.title = "Chart widget"
        super().__init__(data, **kwargs)

    def init_component(self, **kwargs):
        self.register(ChartElement, **kwargs)



import logging
from weakref import ref
from typing import Optional

from ipywidgets import HBox, HTML, Layout, Output, ToggleButton

from cosapp.systems import System
from .sidecar import Sidecar


class WidgetLogger(logging.Handler):
    """Custom logging handler sending logs to an output widget.

    Parameters
    ----------
    system : cosapp.systems.System
        System of interest
    level : int, optional
        Logging level printed; default logging.INFO
    format : str, optional
        Logging formatter string (see `documentation <https://docs.python.org/3/library/logging.html#formatter-objects>`_); default [%(levelname)s] %(message)s
    overwrite: bool, optional
        Remove all other logging handlers specified at root level; default True.
    """

    def __init__(
        self,
        system: System,
        level: int = logging.INFO,
        format: Optional[str] = "[%(levelname)s] %(message)s",
        overwrite: bool = True,
    ):
        super(WidgetLogger, self).__init__(level=level)
        if format:
            self.setFormatter(logging.Formatter(format))

        self._system = ref(system)
        self.auto_clear = True
        self.out = Output(layout=Layout(overflow_y="auto"))

        logger = logging.getLogger()
        if overwrite:
            for handle in logger.handlers:
                logger.removeHandler(handle)
        logger.addHandler(self)

        # Connect events
        system.setup_ran.connect(self._auto_clear)

    def _auto_clear(self):
        """Clear logs if auto_clear is True."""
        if self.auto_clear:
            self.clear_logs()

    def emit(self, record):
        """ Overload of logging.Handler method """
        formatted_record = self.format(record)
        # Workaround for JupyterLab see issue:
        # https://github.com/jupyter-widgets/ipywidgets/issues/1810
        self.out.append_stdout(formatted_record + "\n")
        # with self.out:
        #     print(formatted_record)

    def show_logs(self):
        """ Show the logs """
        auto_clear_but = ToggleButton(
            value=self.auto_clear,
            icon="eraser",
            button_style="",
            tooltip="Toogle log auto clear.",
            layout=Layout(flex="0 0 auto", width="auto"),
        )

        def toogle(change):
            self.auto_clear = change["new"]

        auto_clear_but.observe(toogle, names="value")

        Sidecar(
            [
                HBox(
                    [
                        auto_clear_but,
                        HTML(
                            r"<style>.cosapp-sidecar .jp-OutputPrompt.jp-OutputArea-prompt { display: none;}</style>"
                        ),
                    ],
                    layout=Layout(flex="0 0 auto"),
                ),
                self.out,
            ],
            title="Log",
            anchor="split-bottom",
        )

    def clear_logs(self):
        """ Clear the current logs """
        self.out.clear_output()
        self.out.outputs = ()



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.


from cosapp_lab.widgets.base.base_component import BaseComponent


class DataComponent(BaseComponent):
    name = "DataComponent"

    def computed_notification(self):
        computed_data = self.sys_data.serialize_data_from_system(False)
        recorder_data = self.sys_data.serialize_recorder()
        driver_data = self.sys_data.serialize_driver_data()
        self.time_step = 0
        self.send(
            {
                "type": "DataComponent::update_signal",
                "payload": {
                    "computed_data": computed_data,
                    "recorder_data": recorder_data,
                    "driver_data": driver_data,
                },
            }
        )



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.


from __future__ import print_function
from glob import glob
from os.path import join as pjoin
from pathlib import Path
import os
from setupbase import (
    create_cmdclass,
    get_version,
)
import setuptools

HERE = os.path.abspath(os.path.dirname(__file__))

COSAPP_CONFIG_DIR = Path.home() / ".cosapp.d"

name = "cosapp_lab"

# Get our version
version = get_version(pjoin(name, "_version.py"))
lab_path = os.path.join(HERE, name, "lab_static")
lab_static_path = os.path.join(lab_path, "static")
app_path = os.path.join(HERE, name, "app_static")

# Representative files that should exist after a successful build
jstargets = [
    os.path.join(HERE, "lib", "index.js"),
    os.path.join(HERE, name, "lab_static", "package.json"),
]
package_data_spec = {name: ["*"]}
labext_name = "cosapp_lab"

data_files_spec = [
    ("share/jupyter/labextensions/%s" % labext_name, lab_path, "*.*"),
    ("share/jupyter/labextensions/%s/static" % labext_name, lab_static_path, "*.*"),
    ("etc/jupyter/nbconfig/notebook.d", "jupyter-config", "cosapp_lab.json"),
    (
        "etc/jupyter/jupyter_notebook_config.d",
        "jupyter-config",
        "cosapp_lab_server.json",
    ),
]

cmdclass = create_cmdclass(
    package_data_spec=package_data_spec, data_files_spec=data_files_spec
)

# cmdclass = create_cmdclass(
#     "jsdeps", package_data_spec=package_data_spec, data_files_spec=data_files_spec
# )

# cmdclass["jsdeps"] = combine_commands(
#     install_npm(HERE, build_cmd="build", npm=["jlpm"]),
#     ensure_targets(jstargets)
# )


setup_args = dict(
    scripts=glob(pjoin("scripts", "*")),
    cmdclass=cmdclass,
    entry_points={
        "console_scripts": [
            "cosapp=cosapp_lab.main:cosapp",
        ],
    },
)

if __name__ == "__main__":

    setuptools.setup(**setup_args)
    if not COSAPP_CONFIG_DIR.exists():
        COSAPP_CONFIG_DIR.mkdir(parents=True, exist_ok=True)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.


"""
Widget container for ChartView
"""

from ipywidgets import VBox
from traitlets import Unicode, CaselessStrEnum
from cosapp_lab._frontend import module_name, module_version


class IpyWidgetRender(VBox):
    """Widget container as sidecar panel.

    Attributes
    ----------
    title: str,
        id of div used to attach widget
    children: list of Widget
        List of widget to show in frontend
    """

    _model_name = Unicode("IpyWidgetRenderModel").tag(sync=True)
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)
    _view_name = Unicode("IpyWidgetRenderView").tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)
    title = Unicode("").tag(sync=True)



import os
from typing import Dict, NoReturn, Optional

from cookiecutter.main import cookiecutter
from jinja2 import Environment, FileSystemLoader
from .cosapp_parser import CosappParser
from .utils import replicate_dict_structure
from pathlib import Path


class CosappJsonParser(CosappParser):
    def __init__(self, json_data: Dict):

        super().__init__(json_data)

    def set_up_system(self, data: Dict) -> None:
        self._json_data = data
        self.system_name = list(data["Systems"].keys())[0]
        self._system = None
        try:
            self.system_dict = replicate_dict_structure(data["Systems"])
        except:
            self.system_dict = {}
        self._discover_children(data["Systems"])

    def _discover_children(
        self, system_dict: Dict, parent: Optional[str] = None
    ) -> None:
        """Get the sub system of input system

        Parameters
        ----------
        system : Dict
            The dictionary of input system

        parent: Optional[str]
            Path in form of "parent.child.sub-child ..." to the parent of input system.
            parent = None if input system does not have parent.

        Returns
        -------
        Dict[str, List[str]]
            Dictionary of sub system of input system with their children.
        """
        system_name = list(system_dict.keys())[0]
        system_data = system_dict[system_name]
        if parent is not None:
            key = parent + "." + system_name
        else:
            key = system_name

        if key in self._children:
            content = self._children[key]

        else:
            content = {}

        if "subsystems" not in system_data:
            content["children"] = None
        else:
            content["children"] = list(system_data["subsystems"])

        child_name = ".".join(key.split(".")[1:])

        content["path"] = key
        content["driver_list"] = []
        content["out_port_list"] = list(system_data["outputs"])
        content["in_port_list"] = list(system_data["inputs"])
        content["port_list"] = content["out_port_list"] + content["in_port_list"]
        content["port_data"] = {}

        for port_direction in ["inputs", "outputs"]:
            for port_key, port_val in system_data[port_direction].items():
                if port_key not in content["port_data"]:
                    content["port_data"][port_key] = [
                        _ for _ in list(port_val) if _ != "__class__"
                    ]
                else:
                    content["port_data"][port_key] += [
                        _ for _ in list(port_val) if _ != "__class__"
                    ]

        if key not in self._children:
            self._children[key] = content
        else:
            for content_key in content:
                self._children[key][content_key] = content[content_key]
        if "subsystems" in system_data:
            for sub_key, sub_data in system_data["subsystems"].items():
                self._discover_children({sub_key: sub_data}, key)

    def create_project(self, project_data: Dict = {}, output_dir=".") -> NoReturn:
        """
        Generate cosapp code from json data
        """

        def format_port_data(data: Dict) -> None:
            for key, val in data.items():
                for sub_key, sub_val in val.items():
                    if sub_key in ["unit", "desc"]:
                        val[sub_key] = f'"{sub_val}"'
                    if sub_key in ["limits", "valid_range"] and sub_val is not None:
                        if sub_val[0] == "-inf":
                            sub_val[0] = None
                        if sub_val[1] == "inf":
                            sub_val[1] = None
                        val[sub_key] = tuple(sub_val)

        sys_data = self._json_data
        dir_path = Path(os.path.dirname(__file__)).parent.parent / "utils" / "templates"
        template_path = dir_path / "cookiecutter-cosapp-workspace"
        file_template_path = dir_path / "file_template"
        if len(project_data) > 0:
            cookiecutter(
                template_path,
                no_input=True,
                extra_context=project_data,
                output_dir=output_dir,
                overwrite_if_exists=True,
            )
        else:
            cookiecutter(
                template_path,
                no_input=True,
                output_dir=output_dir,
                overwrite_if_exists=True,
            )
        project_name = project_data.get("project_name", "cosapp_project")
        ports_path = os.path.join(output_dir, project_name, project_name, "ports")
        systems_path = os.path.join(output_dir, project_name, project_name, "systems")

        env = Environment(
            loader=FileSystemLoader(file_template_path),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        port_template = env.get_template("port.j2")
        for class_name, variable_data in sys_data["Ports"].items():
            format_port_data(variable_data)
            port_template.stream(
                class_name=class_name, variable_data=variable_data
            ).dump(os.path.join(ports_path, f"{class_name.lower()}.py"))

        system_template_data = {}
        self.get_cls_definition_from_json(sys_data["Systems"], system_template_data)
        sys_template = env.get_template("system.j2")

        for class_name, variable_data in system_template_data.items():
            variable_data["import"] = {"port": [], "system": []}
            variable_data["module_name"] = project_name
            for direction, extensible_port_name in [
                ("inputs", "inwards"),
                ("outputs", "outwards"),
            ]:
                if direction in variable_data:
                    for port_name in variable_data[direction]:
                        if port_name == extensible_port_name:
                            format_port_data(variable_data[direction][port_name])
                        else:
                            port_cls_name = variable_data[direction][port_name][
                                "__class__"
                            ]
                            if port_cls_name not in variable_data["import"]["port"]:
                                variable_data["import"]["port"].append(port_cls_name)
            if "subsystems" in variable_data:
                for sys_cls_name in variable_data["subsystems"].values():
                    real_name = sys_cls_name.replace("__main__.", "")
                    if real_name not in variable_data["import"]["system"]:
                        variable_data["import"]["system"].append(real_name)

            sys_template.stream(class_name=class_name, **variable_data).dump(
                os.path.join(systems_path, f"{class_name.lower()}.py")
            )

    @classmethod
    def get_cls_definition_from_json(cls, input: Dict, output: Dict) -> None:
        data = input[list(input.keys())[0]]
        class_name = data["class"].replace("__main__.", "")
        if class_name not in output:
            output[class_name] = {"subsystems": {}}
            for key, value in data.items():
                if key == "subsystems":
                    for sub_key, sub_val in value.items():
                        output[class_name]["subsystems"][sub_key] = sub_val["class"]
                elif key == "class":
                    pass
                else:
                    output[class_name][key] = value
        if "subsystems" in data:
            for system_key, system_data in data["subsystems"].items():
                cls.get_cls_definition_from_json({system_key: system_data}, output)
